# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import enum
import types
import typing
from typing import Set  # noqa: F401 # imported but unused (used in exec() context)
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Final,
    ForwardRef,
    Generic,
    List,
    Literal,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import devtools
import factory
import pytest
import pytest_factoryboy as pytfboy

from gt4py.eve import datamodels, utils


T = TypeVar("T")
U = TypeVar("U")


SAMPLE_MODEL_FACTORIES: List[factory.Factory] = []


def register_factory(
    factory_class: Optional[factory.Factory] = None,
    *,
    model_fixture: Optional[str] = None,
    collection: Optional[MutableSequence[factory.Factory]] = None,
) -> Union[factory.Factory, Callable[[factory.Factory], factory.Factory]]:
    def _decorator(factory: factory.Factory) -> factory.Factory:
        if collection is not None:
            collection.append(factory)
        return pytfboy.register(factory, model_fixture)

    return _decorator(factory_class) if factory_class is not None else _decorator


class SampleEnum(enum.Enum):
    FOO = "foo"
    BLA = "bla"


class SampleClass:
    pass


class EmptyModel(datamodels.DataModel):
    pass


@register_factory(collection=SAMPLE_MODEL_FACTORIES)
class EmptyModelFactory(factory.Factory):
    class Meta:
        model = EmptyModel


class IntModel(datamodels.DataModel):
    value: int


@register_factory(collection=SAMPLE_MODEL_FACTORIES)
class IntModelFactory(factory.Factory):
    class Meta:
        model = IntModel

    value = 1


class AnyModel(datamodels.DataModel):
    value: Any


@register_factory(collection=SAMPLE_MODEL_FACTORIES)
class AnyModelFactory(factory.Factory):
    class Meta:
        model = AnyModel

    value = ("any", "value")


class GenericModel(datamodels.GenericDataModel, Generic[T]):
    value: T


@register_factory(collection=SAMPLE_MODEL_FACTORIES)
class GenericModelFactory(factory.Factory):
    class Meta:
        model = GenericModel

    value = "generic value"


@pytest.fixture(params=SAMPLE_MODEL_FACTORIES)
def example_model_factory(request) -> datamodels.DataModel:
    return request.param


# --- Tests ---
# Test generated members
def test_datamodel_class_members(example_model_factory):
    model = example_model_factory()
    model_class = model.__class__

    assert hasattr(model_class, "__init__")
    assert hasattr(model_class, "__datamodel_fields__")
    assert isinstance(model_class.__datamodel_fields__, utils.FrozenNamespace)
    assert hasattr(model_class, "__datamodel_params__")
    assert isinstance(model_class.__datamodel_params__, utils.FrozenNamespace)
    assert hasattr(model_class, "__datamodel_root_validators__")
    assert isinstance(model_class.__datamodel_root_validators__, tuple)


def test_devtools_compatibility(example_model_factory):
    model = example_model_factory()
    model_class = model.__class__
    formatted_string = devtools.pformat(model)

    assert hasattr(model_class, "__pretty__")
    assert callable(model_class.__pretty__)
    assert f"{model_class.__name__}(" in formatted_string
    for name in model_class.__datamodel_fields__.keys():
        assert f"{name}=" in formatted_string


class TestInitialization:
    def test_init(self):
        @datamodels.datamodel
        class Model:
            value: int
            enum_value: SampleEnum
            list_value: List[int]

        model = Model(value=1, enum_value=SampleEnum.FOO, list_value=[1, 2, 3])
        assert model.value == 1
        assert model.enum_value == SampleEnum.FOO
        assert model.list_value == [1, 2, 3]

    def test_custom_init(self):
        class ModelWithCustomInit(datamodels.DataModel):
            value: float
            enum_value: SampleEnum
            list_value: List[float]

            def __init__(self, single_value: float) -> None:
                self.__auto_init__(single_value, SampleEnum.BLA, [1.0] * int(single_value))

        model = ModelWithCustomInit(3.5)
        assert model.value == 3.5
        assert model.enum_value == SampleEnum.BLA
        assert model.list_value == [1.0, 1.0, 1.0]

    def test_init_hooks(self):
        @datamodels.datamodel
        class ModelWithInitHooks:
            value: float
            STATIC_INT: ClassVar[int] = 0

            def __pre_init__(self) -> None:
                self.__class__.STATIC_INT += 1

            def __post_init__(self) -> None:
                self.value *= 10

        assert ModelWithInitHooks.STATIC_INT == 0
        model = ModelWithInitHooks(3.5)
        assert ModelWithInitHooks.STATIC_INT == 1
        assert model.value == 3.5 * 10

    def test_custom_init_and_hooks(self):
        class ModelWithCustomInitAndHooks(datamodels.DataModel):
            value: float
            STATIC_INT: ClassVar[int] = 0

            def __init__(self, str_value: str) -> None:
                self.__auto_init__(float(str_value))

            def __pre_init__(self) -> None:  # type: ignore[override]
                self.__class__.STATIC_INT += 1

            def __post_init__(self) -> None:  # type: ignore[override]
                self.value *= 10

        assert ModelWithCustomInitAndHooks.STATIC_INT == 0
        model = ModelWithCustomInitAndHooks("-5.25")
        assert ModelWithCustomInitAndHooks.STATIC_INT == 1
        assert model.value == -5.25 * 10

    def test_non_datamodels_init(self):
        @datamodels.datamodel
        class ModelWithReallyCustomInit:
            value: float
            STATIC_INT: ClassVar[int] = 0

            def __init__(self, str_value: str) -> None:
                self.value = float(str_value)

            def __pre_init__(self) -> None:
                self.__class__.STATIC_INT += 1

            def __post_init__(self) -> None:
                self.value *= 10

        assert ModelWithReallyCustomInit.STATIC_INT == 0
        model = ModelWithReallyCustomInit("-5.25")
        assert ModelWithReallyCustomInit.STATIC_INT == 0
        assert model.value == -5.25


def test_kw_only_model():
    @datamodels.datamodel
    class Model:
        int_value: int
        str_value: str

    @datamodels.datamodel(kw_only=True)
    class KwModel:
        int_value: int
        str_value: str

    @datamodels.datamodel(kw_only=False)
    class NotKwModel:
        int_value: int
        str_value: str

    model = Model(1, "foo")
    kw_model = KwModel(int_value=1, str_value="foo")
    not_kw_model = NotKwModel(1, str_value="foo")
    assert (model.int_value, model.str_value) == (kw_model.int_value, kw_model.str_value)
    assert (model.int_value, model.str_value) == (not_kw_model.int_value, not_kw_model.str_value)

    with pytest.raises(TypeError, match="takes 1 positional argument but 3 were given"):
        KwModel(1, "foo")

    with pytest.raises(TypeError, match="takes 1 positional argument but 2 positional"):
        KwModel(1, str_value="foo")

    @datamodels.datamodel
    class MixedModel:
        int_value: int
        str_value: str = datamodels.field(kw_only=True)

    assert MixedModel(int_value=3, str_value="foo") == MixedModel(3, str_value="foo")

    with pytest.raises(TypeError, match="takes 2 positional arguments but 3 were given"):
        MixedModel(3, "foo")


def test_slots():
    class Model:
        as_int: int
        a_str: str

    assert "__slots__" not in Model.__dict__

    SlottedModel = datamodels.datamodel(Model, slots=True)
    model = SlottedModel(33, "foo")

    assert (model.as_int, model.a_str) == (33, "foo")
    assert "__dict__" not in dir(model)
    assert "__slots__" in SlottedModel.__dict__
    assert SlottedModel is not Model

    DataModel = datamodels.datamodel(Model, slots=False)
    model = DataModel(33, "foo")

    assert (model.as_int, model.a_str) == (33, "foo")
    assert "__dict__" in dir(model)
    assert "__slots__" not in DataModel.__dict__
    assert DataModel is Model


class TestConversion:
    def test_simple_coertion_in_decorator(self):
        @datamodels.datamodel(coerce=True)
        class CoercedModel:
            as_int: int
            a_str: str

        instance = CoercedModel(-2, "2")
        assert instance.as_int == -2
        assert instance.a_str == "2"

        instance = CoercedModel(-2, 2)
        assert instance.as_int == -2
        assert instance.a_str == "2"

        A_STR = """
            Lorem ipsum dolor sit amet, nonumy accusam suscipit et mei, ipsum saperet no nec,
            te volumus insolens nam. Verear scripserit delicatissimi cu vis, eam graeci facete in.
            Atqui inani maiorum sea ex. Vim vidit intellegam eu. Mei dico lorem eu, at per paulo
            aperiri admodum. Summo iriure consequuntur per ea, his ex amet tacimates.
        """
        instance = CoercedModel("-2", A_STR)
        assert instance.as_int == -2
        assert instance.a_str == A_STR
        assert instance.a_str is A_STR

    def test_simple_coertion_in_fields(self):
        class PartiallyCoercedModel(datamodels.DataModel):
            as_int: int = datamodels.field(converter="coerce")
            only_int: int

        instance = PartiallyCoercedModel(-2, 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        instance = PartiallyCoercedModel("-2", 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        with pytest.raises(TypeError, match="only_int"):
            PartiallyCoercedModel("-2", "2")

        class PartiallyCoercedModel(datamodels.DataModel):
            as_int: datamodels.Coerced[int]
            only_int: int

        instance = PartiallyCoercedModel(-2, 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        instance = PartiallyCoercedModel("-2", 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        with pytest.raises(TypeError, match="only_int"):
            PartiallyCoercedModel("-2", "2")

    def test_custom_coertion_in_fields(self):
        class CustomCoercedModel(datamodels.DataModel):
            as_int: int = datamodels.field(converter=int)
            only_int: int

        instance = CustomCoercedModel(-2, 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        instance = CustomCoercedModel("-2", 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        with pytest.raises(TypeError, match="only_int"):
            CustomCoercedModel("-2", "2")

        with pytest.raises(
            TypeError, match="'CustomCoercedModel.as_int' which already defines a custom converter"
        ):

            class CustomCoercedModel(datamodels.DataModel):
                as_int: datamodels.Coerced[int] = datamodels.field(converter=int)
                only_int: int

    def test_coertion_of_collections(self):
        @datamodels.datamodel(coerce=True)
        class ModelWithCollections:
            items: List[int]
            table: Dict[int, Tuple[str, str]]

        instance = ModelWithCollections([], {})
        assert instance.items == []
        assert instance.table == {}

        instance = ModelWithCollections([1, 2], {2: ("two", "zwei")})
        assert instance.items == [1, 2]
        assert instance.table == {2: ("two", "zwei")}

        instance = ModelWithCollections((-1, -2), [(2, ("two", "zwei")), (3, ("three", "drei"))])
        assert instance.items == [-1, -2]
        assert instance.table == {2: ("two", "zwei"), 3: ("three", "drei")}

        with pytest.raises(TypeError, match="items"):
            instance = ModelWithCollections(["1", "2"], {})
        with pytest.raises(TypeError, match="items"):
            instance = ModelWithCollections(("1", "2"), {})


class TestAnnotationMarkers:
    def test_coerced_marker(self):
        @datamodels.datamodel
        class CoercedModel:
            as_int: datamodels.Coerced[int]
            only_str: str

        instance = CoercedModel(-2, "2")
        assert instance.as_int == -2
        assert instance.only_str == "2"

        instance = CoercedModel("-2", "2")
        assert instance.as_int == -2
        assert instance.only_str == "2"

        with pytest.raises(TypeError, match="only_str"):
            CoercedModel("-2", 2)

    def test_unchecked_marker(self):
        @datamodels.datamodel
        class UncheckedModel:
            as_int: datamodels.Unchecked[int]
            only_int: int

        instance = UncheckedModel(-2, 2)
        assert instance.as_int == -2
        assert instance.only_int == 2

        instance = UncheckedModel(-2.5, 2)
        assert instance.as_int == -2.5
        assert instance.only_int == 2

        with pytest.raises(TypeError, match="only_int"):
            instance = UncheckedModel(-2.5, 2.5)

    def test_combined_markers(self):
        @datamodels.datamodel
        class UncheckedAndCoercedModel:
            a: datamodels.Unchecked[datamodels.Coerced[int]]
            b: datamodels.Coerced[datamodels.Unchecked[int]]

        instance = UncheckedAndCoercedModel(-2.5, "-2")
        assert instance.a == -2 == instance.b

        instance = UncheckedAndCoercedModel("-2", -2.5)
        assert instance.a == -2 == instance.b


class TestTypeValidationFactory:
    def test_no_factory(self):
        @datamodels.datamodel(type_validation_factory=None)
        class UnsafeModel:
            an_int: int
            a_float: float

        model = UnsafeModel("foo", "bar")
        assert isinstance(model, UnsafeModel)
        assert model.an_int, model.a_float == ("foo", "bar")

    def test_custom_type_validation_factory(self):
        def custom_factory(type_hint, name):
            if type_hint is int:

                def _validator(model, attrib, value):
                    if value % 2:
                        raise TypeError("FOO")

                return _validator

            else:
                return lambda model, attrib, value: None

        @datamodels.datamodel(type_validation_factory=custom_factory)
        class BooModel:
            an_int: int
            a_float: float

        model = BooModel(2, 2.0)
        assert model.an_int, model.a_float == (2, 2.0)

        model = BooModel(2, "a float")
        assert model.an_int, model.a_float == (2, "a float")

        with pytest.raises(TypeError, match="FOO"):
            model = BooModel(3, 2.0)


def test_default_values():
    @datamodels.datamodel
    class Model:
        bool_value: bool
        int_value: int = 1
        enum_value: SampleEnum = SampleEnum.FOO
        any_value: Any = datamodels.field(default="ANY")

    model = Model(False)
    with pytest.raises(TypeError, match="'bool_value'"):
        Model()

    assert model.bool_value is False
    assert model.int_value == 1
    assert model.enum_value == SampleEnum.FOO
    assert model.any_value == "ANY"

    @datamodels.datamodel
    class WrongModel:
        bool_value: bool = 1

    with pytest.raises(TypeError, match="'WrongModel.bool_value'"):
        WrongModel()


def test_default_factories():
    @datamodels.datamodel
    class Model:
        list_value: List[int] = datamodels.field(default_factory=list)
        dict_value: Dict[str, int] = datamodels.field(default_factory=lambda: {"one": 1})

    model = Model()

    assert model.list_value == []
    assert model.dict_value == {"one": 1}

    @datamodels.datamodel
    class WrongModel:
        list_value: List[int] = datamodels.field(default_factory=tuple)

    with pytest.raises(TypeError, match="'WrongModel.list_value'"):
        WrongModel()


# Test field specification
SAMPLE_TYPE_DATA: Final = [
    ("bool", [True, False], [1, "True"]),
    ("int", [1, -1], [1.0, True, "1"]),
    ("float", [1.0], [1, "1.0"]),
    ("str", ["", "one"], [1, ("one",)]),
    ("complex", [1j], [1, 1.0, "1j"]),
    ("bytes", [b"bytes", b""], ["string", ["a"]]),
    ("typing.Any", ["any"], tuple()),
    ("typing.Literal[1, 1.0, True]", [1, 1.0, True], [False]),
    ("typing.Tuple[int, str]", [(3, "three")], [(), (3, 3)]),
    ("typing.Tuple[int, ...]", [(1, 2, 3), ()], [3, (3, "three")]),
    ("typing.List[int]", ([1, 2, 3], []), (1, [1.0])),
    ("typing.Set[int]", ({1, 2, 3}, set()), (1, [1], (1,), {1: None})),
    ("typing.Dict[int, str]", ({}, {3: "three"}), ([(3, "three")], 3, "three", [])),
    ("typing.Sequence[int]", ([1, 2, 3], [], (1, 2, 3), tuple()), (1, [1.0], {1})),
    ("typing.MutableSequence[int]", ([1, 2, 3], []), ((1, 2, 3), tuple(), 1, [1.0], {1})),
    ("typing.Set[int]", ({1, 2, 3}, set()), (1, [1], (1,), {1: None})),
    ("typing.Union[int, float, str]", [1, 3.0, "one"], [[1], [], 1j]),
    ("typing.Optional[int]", [1, None], [[1], [], 1j]),
    (
        "typing.Dict[Union[int, float, str], Union[Tuple[int, Optional[float]], Set[int]]]",
        [{1: (2, 3.0)}, {1.0: (2, None)}, {"1": {1, 2}}],
        [{(1, 1.0, "1"): set()}, {1: [1]}, {"1": (1,)}],
    ),
]


@pytest.mark.parametrize(["type_hint", "valid_values", "wrong_values"], SAMPLE_TYPE_DATA)
def test_field_type_hint(type_hint: str, valid_values: Sequence[Any], wrong_values: Sequence[Any]):
    context: Dict[str, Any] = {}
    exec(
        f"""
@datamodels.datamodel
class Model:
    value: {type_hint}
""",
        globals(),
        context,
    )
    Model = context["Model"]

    field_type = Model.__datamodel_fields__.value.type
    if isinstance(field_type, str):
        field_type = eval(field_type)
    assert field_type == eval(type_hint)

    for value in valid_values:
        Model(value=value)

    for value in wrong_values:
        with pytest.raises((TypeError, ValueError), match="'Model.value'"):
            Model(value=value)


def test_custom_class_type_hint():
    @datamodels.datamodel
    class Model1:
        value: SampleEnum

    Model1(value=SampleEnum.FOO)

    class SampleClassChild(SampleClass):
        pass

    class SampleClassGrandChild(SampleClassChild):
        pass

    @datamodels.datamodel
    class Model2:
        value: SampleClass

    Model2(value=SampleClass())
    Model2(value=SampleClassChild())
    Model2(value=SampleClassGrandChild())

    class AnotherClass:
        pass

    with pytest.raises(TypeError, match="value"):
        Model1(value=AnotherClass())
    with pytest.raises(TypeError, match="value"):
        Model2(value=AnotherClass())


class MyType:
    def __init__(self, value: Any) -> None:
        self.value = value

    def add(self, something) -> None:
        return self.value + something

    @classmethod
    def __type_validator__(cls) -> datamodels.FieldValidator:
        def _custom_validator(
            instance: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
        ) -> None:
            if not (hasattr(value, "value") and hasattr(value, "add")):
                raise TypeError("Invalid value type for '{attribute.name}' field.")

        return _custom_validator


@datamodels.datamodel
class GlobalRecursiveModel:
    value: Optional[GlobalRecursiveModel] = None
    others: Optional[Dict[Literal["A", "B"], Union[str, GlobalRecursiveModel]]] = None


def test_deferred_class_type_hint():
    # Model defined in a module global context
    assert isinstance(GlobalRecursiveModel.__datamodel_fields__.value.type, ForwardRef)
    assert isinstance(GlobalRecursiveModel.__datamodel_fields__.others.type, ForwardRef)

    m1 = GlobalRecursiveModel()
    m2 = GlobalRecursiveModel(value=m1)
    m3 = GlobalRecursiveModel(value=m2)
    GlobalRecursiveModel(value=m2, others={"A": m3, "B": "something"})

    with pytest.raises(TypeError, match="value"):
        GlobalRecursiveModel(value="wrong_value")
    with pytest.raises(TypeError, match="others"):
        GlobalRecursiveModel(others={"A": -1})
    with pytest.raises(TypeError, match="others"):
        GlobalRecursiveModel(others={"a": "wrong"})

    assert GlobalRecursiveModel.__datamodel_fields__.value.type.__args__[0] == GlobalRecursiveModel
    assert (
        GlobalRecursiveModel.__datamodel_fields__.others.type.__args__[0].__args__[1].__args__[1]
        == GlobalRecursiveModel
    )

    # Models defined in a non-global context
    @datamodels.datamodel
    class RecursiveModel:
        int_value: int
        list_value: List[RecursiveModel]

    assert isinstance(RecursiveModel.__datamodel_fields__.list_value.type, ForwardRef)
    datamodels.update_forward_refs(RecursiveModel, {"RecursiveModel": RecursiveModel})
    assert RecursiveModel.__datamodel_fields__.list_value.type.__args__[0] == RecursiveModel

    m1 = RecursiveModel(int_value=1, list_value=[])
    m2 = RecursiveModel(int_value=1, list_value=[m1])
    RecursiveModel(int_value=1, list_value=[m1, m2])

    with pytest.raises(TypeError, match="list_value"):
        RecursiveModel(int_value=1, list_value=["wrong_value"])

    @datamodels.datamodel
    class CollectorModel:
        value1: NotYetDefinedModel1
        value2: NotYetDefinedModel2

    @datamodels.datamodel
    class NotYetDefinedModel1:
        int_value: int

    @datamodels.datamodel
    class NotYetDefinedModel2:
        int_value: int

    assert isinstance(CollectorModel.__datamodel_fields__.value1.type, ForwardRef)
    assert isinstance(CollectorModel.__datamodel_fields__.value2.type, ForwardRef)
    datamodels.update_forward_refs(
        CollectorModel,
        {"NotYetDefinedModel1": NotYetDefinedModel1, "NotYetDefinedModel2": NotYetDefinedModel2},
    )
    assert CollectorModel.__datamodel_fields__.value1.type == NotYetDefinedModel1
    assert CollectorModel.__datamodel_fields__.value2.type == NotYetDefinedModel2

    CollectorModel(value1=NotYetDefinedModel1(int_value=1), value2=NotYetDefinedModel2(int_value=2))
    with pytest.raises(TypeError, match="value2"):
        CollectorModel(value1=NotYetDefinedModel1(int_value=1), value2=2)


def test_field_redefinition():
    class Model(datamodels.DataModel):
        value: int = 0

    assert Model().value == 0

    # Redefinition with same type
    class ChildModel1(Model):
        value: int = 1

    assert ChildModel1().value == 1

    # Redefinition with different type
    class ChildModel2(Model):
        value: float = 2.2  # type: ignore  # redefining value as float on purpose

    assert ChildModel2().value == 2.2

    with pytest.raises(TypeError, match="value"):
        assert ChildModel2(value=2)


def test_class_vars():
    @datamodels.datamodel
    class Model:
        value: Any
        default_value: Any = None

        class_var: ClassVar[int] = 0

    field_names = set(Model.__datamodel_fields__.keys())
    assert field_names == {"value", "default_value"}
    assert hasattr(Model, "class_var")
    assert Model.class_var == 0


def test_missing_annotations():
    with pytest.raises(TypeError, match="other_value"):

        class Model(datamodels.DataModel):
            other_value = datamodels.field(default=None)


# Test custom validators
class ModelWithValidators(datamodels.DataModel):
    bool_value: bool = False
    int_value: int = 0
    even_int_value: int = 2
    float_value: float = 0.0
    str_value: str = ""
    extra_value: Optional[Any] = None

    @datamodels.validator("bool_value")
    def _bool_value_validator(
        self: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
    ) -> None:
        assert isinstance(self, ModelWithValidators)

    @datamodels.validator("int_value")
    def _int_value_validator(
        self: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
    ) -> None:
        if value < 0:
            raise ValueError(f"'{attribute.name}' must be larger or equal to 0")

    @datamodels.validator("even_int_value")
    def _even_int_value_validator(
        self: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
    ) -> None:
        if value % 2:
            raise ValueError(f"'{attribute.name}' must be an even number")

    @datamodels.validator("float_value")
    def _float_value_validator(
        self: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
    ) -> None:
        if value > 3.14159:
            raise ValueError(f"'{attribute.name}' must be lower or equal to 3.14159")

    @datamodels.validator("str_value")
    def _str_value_validator(
        self: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
    ) -> None:
        # This kind of validation should arguably happen in a root_validator, but
        # since float_value is defined before str_value, it should have been
        # already validated at this point
        if value == str(cast(ModelWithValidators, self).float_value):
            raise ValueError(f"'{attribute.name}' must be different to 'float_value'")

    @datamodels.validator("extra_value")
    def _extra_value_validator(
        self: datamodels.DataModel, attribute: datamodels.Attribute, value: Any
    ) -> None:
        if bool(value):
            raise ValueError(f"'{attribute.name}' must be equivalent to False")


class ChildModelWithValidators(ModelWithValidators):
    pass


@typing.no_type_check
@pytest.mark.parametrize("model_class", [ModelWithValidators, ChildModelWithValidators])
def test_field_validators(model_class: Type[Union[ModelWithValidators, ChildModelWithValidators]]):

    with pytest.raises(ValueError, match="int_value"):
        model_class(int_value=-1)

    model_class(even_int_value=-2)
    with pytest.raises(ValueError, match="even_int_value"):
        model_class(even_int_value=1)

    model_class(float_value=-2.0)
    with pytest.raises(ValueError, match="float_value"):
        model_class(float_value=4.0)

    with pytest.raises(ValueError, match="str_value"):
        model_class(float_value=1.0, str_value="1.0")

    with pytest.raises(ValueError, match="extra_value"):
        model_class(extra_value=1)

    model_class(extra_value=False)
    model_class(extra_value=0)
    model_class(extra_value="")
    model_class(extra_value=[])
    with pytest.raises(ValueError, match="extra_value"):
        model_class(extra_value=1)


def test_new_field_validators_in_subclass():
    class ChildModelWithValidators(ModelWithValidators):
        @datamodels.validator("int_value")
        def _int_value_validator(self, attribute, value) -> None:
            if value > 10:
                raise ValueError(f"'{attribute.name}' must be lower or equal to 10")

    ChildModelWithValidators(int_value=1)
    with pytest.raises(ValueError, match="int_value"):
        ChildModelWithValidators(int_value=-1)
    with pytest.raises(ValueError, match="int_value"):
        ChildModelWithValidators(int_value=11)


def test_field_validators_in_overwritten_field_in_subclass():
    class ChildModelWithValidators(ModelWithValidators):
        int_value: int = 0

        @datamodels.validator("int_value")
        def _int_value_validator(self, attribute, value) -> None:
            if value > 10:
                raise ValueError(f"'{attribute.name}' must be lower or equal to 10")

    ChildModelWithValidators(int_value=1)
    ChildModelWithValidators(int_value=-1)  # Test new field definition
    with pytest.raises(ValueError, match="int_value"):
        ChildModelWithValidators(int_value=11)

    # Overwrite field definition with a new type
    class AnotherChildModelWithValidators(ModelWithValidators):
        extra_value: float = 0.0

        @datamodels.validator("extra_value")
        def _extra_value_validator(self, attribute, value) -> None:
            if value < 0.0:
                raise ValueError(f"'{attribute.name}' must be a positive number")

    AnotherChildModelWithValidators(extra_value=0.0)
    AnotherChildModelWithValidators(extra_value=1.0)
    with pytest.raises(ValueError, match="extra_value"):
        AnotherChildModelWithValidators(extra_value=-1.0)


class ModelWithRootValidators(datamodels.DataModel):
    int_value: int
    float_value: float
    str_value: str

    class_counter: ClassVar[int] = 0

    @datamodels.root_validator
    @classmethod
    def _root_validator(cls: Type[datamodels.DataModel], instance: datamodels.DataModel) -> None:
        assert cls is type(instance)
        assert issubclass(cls, ModelWithRootValidators)
        assert isinstance(instance, ModelWithRootValidators)
        cls.class_counter = 0

    @datamodels.root_validator
    @classmethod
    def _another_root_validator(
        cls: Type[datamodels.DataModel], instance: datamodels.DataModel
    ) -> None:
        cls = cast(Type[ModelWithRootValidators], cls)
        instance = cast(ModelWithRootValidators, instance)
        assert cls.class_counter == 0
        cls.class_counter += 1

    @datamodels.root_validator
    @classmethod
    def _final_root_validator(
        cls: Type[datamodels.DataModel], instance: datamodels.DataModel
    ) -> None:
        cls = cast(Type[ModelWithRootValidators], cls)
        instance = cast(ModelWithRootValidators, instance)
        assert cls.class_counter == 1
        cls.class_counter += 1

        if instance.int_value == instance.float_value:
            raise ValueError("'int_value' and 'float_value' must be different")


class ChildModelWithRootValidators(ModelWithRootValidators):
    pass


@pytest.mark.parametrize("model_class", [ModelWithRootValidators, ChildModelWithRootValidators])
def test_root_validators(model_class: Type[datamodels.DataModel]):
    model_class(int_value=0, float_value=1.1, str_value="")
    with pytest.raises(ValueError, match="float_value"):
        model_class(int_value=1, float_value=1.0, str_value="")


def test_root_validators_in_subclasses():
    class Model(ModelWithRootValidators):
        @datamodels.root_validator
        @classmethod
        def _root_validator(cls, instance):
            assert cls.class_counter == 2
            cls.class_counter += 10

        @datamodels.root_validator
        @classmethod
        def _another_root_validator(cls, instance):
            assert cls.class_counter == 12
            cls.class_counter += 10

        @datamodels.root_validator
        @classmethod
        def _final_root_validator(cls, instance):
            assert cls.class_counter == 22
            if str(instance.int_value) == instance.str_value:
                raise ValueError("'int_value' and 'str_value' must be different")

    with pytest.raises(ValueError, match="str_value"):
        Model(int_value=1, float_value=0.0, str_value="1")
    with pytest.raises(ValueError, match="float_value"):
        Model(int_value=1, float_value=1.0, str_value="1")
    with pytest.raises(ValueError, match="float_value"):
        Model(int_value=1, float_value=1.0, str_value="1")


# Test field options
def test_field_init():
    @datamodels.datamodel
    class Model:
        hidden_value: int = datamodels.field(init=False)
        value: int
        hidden_value_with_default: int = datamodels.field(default=10, init=False)

    model = Model(value=1)
    assert model.value == 1
    assert model.hidden_value_with_default == 10

    model.hidden_value = -1
    assert model.hidden_value == -1


def test_field_metadata():
    @datamodels.datamodel
    class Model:
        value: int = datamodels.field(metadata={"my_metadata": "META"})

    assert isinstance(Model.__datamodel_fields__.value.metadata, types.MappingProxyType)
    assert Model.__datamodel_fields__.value.metadata["my_metadata"] == "META"


# Test datamodel options
class TestDatamodelOptions:
    def test_frozen(self):
        import attr  # type: ignore[import] # Missing library stubs for Python 3.10)

        @datamodels.datamodel(frozen=True)
        class FrozenModel:
            value: Any = None

        assert FrozenModel.__datamodel_params__.frozen is True
        with pytest.raises(attr.exceptions.FrozenInstanceError):
            FrozenModel().value = 1

        class FrozenModel2(datamodels.DataModel, frozen=True):
            value: Any = None

        assert FrozenModel2.__datamodel_params__.frozen is True
        with pytest.raises(attr.exceptions.FrozenInstanceError):
            FrozenModel2().value = 1

    # Explanation of expected hash() behavior in Python:
    #   https://hynek.me/articles/hashes-and-equality/
    def test_safe_mutable_hash(self):
        string_value = "this is a really long string to avoid string interning 1234567890 +:?.!"

        # Safe mutable case: no hash
        @datamodels.datamodel
        class MutableModel:
            value: Any = None

        with pytest.raises(TypeError, match="unhashable type"):
            hash(MutableModel(value=string_value))

    def test_unsafe_mutable_hash(self):
        # Unsafe mutable case: mutable hash
        @datamodels.datamodel(unsafe_hash=True)
        class MutableModelWithHash:
            value: Any = None

        string_value = "this is a really long string to avoid string interning 1234567890 +:?.!"
        mutable_model = MutableModelWithHash(value=string_value)
        hash_value = hash(mutable_model)
        assert hash_value == hash(MutableModelWithHash(value=string_value))

        # This is the (expected) wrong behaviour with 'unsafe_hash=True',
        # because hash value should not change during lifetime of an object
        mutable_model.value = 42
        assert hash_value != hash(mutable_model)

    def test_safe_frozen_hash(self):
        # Safe frozen case: proper hash
        @datamodels.datamodel(frozen=True)
        class FrozenModel:
            value: Any = None

        string_value = "this is a really long string to avoid string interning 1234567890 +:?.!"

        assert hash(FrozenModel(value=string_value)) == hash(FrozenModel(value=string_value))


# Test module functions
def test_info_functions():
    @datamodels.datamodel
    class Model:
        int_value: int
        float_value: float
        str_value: str
        class_counter: ClassVar[int] = 0

    fields_info = datamodels.fields(Model)
    fields_info_keys = set(fields_info.keys())

    assert isinstance(fields_info, utils.FrozenNamespace)
    assert fields_info_keys == {"int_value", "float_value", "str_value"}
    assert datamodels.get_fields(Model) == fields_info

    model = Model(int_value=1, float_value=2.0, str_value="string")

    assert fields_info == datamodels.fields(model)
    assert datamodels.asdict(model) == {"int_value": 1, "float_value": 2.0, "str_value": "string"}
    assert datamodels.astuple(model) == (1, 2.0, "string")


# Test generic models
CONCRETE_TYPE_SAMPLES: Final = [int, List[float], Tuple[int, ...], Optional[int], Union[int, float]]


@pytest.mark.parametrize("concrete_type", CONCRETE_TYPE_SAMPLES)
def test_generic_model_instantiation_name(concrete_type: Type):
    Model: Type[datamodels.DataModel] = datamodels.concretize(GenericModel, concrete_type)  # type: ignore[misc]
    assert Model.__name__.startswith(GenericModel.__name__)
    assert Model.__name__ != GenericModel.__name__

    Model = datamodels.concretize(GenericModel, concrete_type, class_name="MyNewConcreteClass")  # type: ignore[misc]
    assert Model.__name__ == "MyNewConcreteClass"


@pytest.mark.parametrize("concrete_type", CONCRETE_TYPE_SAMPLES)
def test_generic_model_alias(concrete_type: Type):
    Model: Type[datamodels.DataModel] = datamodels.concretize(GenericModel, concrete_type)  # type: ignore[misc]

    class SubModel(GenericModel[concrete_type]):  # type: ignore  # using run-time type on purpose
        ...

    assert SubModel.__base__ is Model


@pytest.mark.parametrize("concrete_type", CONCRETE_TYPE_SAMPLES)
def test_generic_model_instantiation_cache(concrete_type):
    Model1 = datamodels.concretize(GenericModel, concrete_type)
    Model2 = datamodels.concretize(GenericModel, concrete_type)
    Model3 = datamodels.concretize(GenericModel, concrete_type)

    assert (
        Model1 is Model2
        and Model2 is Model3
        and Model3 is datamodels.concretize(GenericModel, concrete_type)
    )


class TestGenericModelValidation:
    def test_basic_generic_field(self):
        class GenericModel(datamodels.GenericDataModel, Generic[T]):
            value: T

        GenericModel(value=1)
        GenericModel(value="value")
        GenericModel(value=(1.0, "value"))
        GenericModel(value=None)

    def test_partial_generic_field(self):
        class PartialGenericModel(datamodels.GenericDataModel, Generic[T]):
            value: List[T]

        PartialGenericModel(value=[])
        PartialGenericModel(value=[1])
        PartialGenericModel(value=["value"])
        PartialGenericModel(value=[1.0, "value"])
        PartialGenericModel(value=[(1.0, "value")])
        PartialGenericModel(value=[None])
        with pytest.raises(TypeError, match="'PartialGenericModel.value'"):
            PartialGenericModel(value=1)
        with pytest.raises(TypeError, match="'PartialGenericModel.value'"):
            PartialGenericModel(value=(1, 2))

        PartialGenericModel__int = PartialGenericModel[int]
        PartialGenericModel__int(value=[])
        PartialGenericModel__int(value=[1])
        with pytest.raises(TypeError, match="'PartialGenericModel__int.value'"):
            PartialGenericModel__int(value=1)
        with pytest.raises(TypeError, match="'PartialGenericModel__int.value'"):
            PartialGenericModel__int(value=(1, 2))
        with pytest.raises(TypeError, match="'PartialGenericModel__int.value'"):
            PartialGenericModel__int(value=[1.0])
        with pytest.raises(TypeError, match="'PartialGenericModel__int.value'"):
            PartialGenericModel__int(value=["1"])

    def test_partial_specialization(self):
        class PartialGenericModel(datamodels.GenericDataModel, Generic[T, U]):
            value: List[Tuple[T, U]]

        PartialGenericModel(value=[])
        PartialGenericModel(value=[("value", 3)])
        PartialGenericModel(value=[(1, "value")])
        PartialGenericModel(value=[(-1.0, "value")])
        with pytest.raises(TypeError, match="'PartialGenericModel.value'"):
            PartialGenericModel(value=1)
        with pytest.raises(TypeError, match="'PartialGenericModel.value'"):
            PartialGenericModel(value=(1, 2))
        with pytest.raises(TypeError, match="'PartialGenericModel.value'"):
            PartialGenericModel(value=[()])
        with pytest.raises(TypeError, match="'PartialGenericModel.value'"):
            PartialGenericModel(value=[(1,)])

        print(f"{PartialGenericModel.__parameters__=}")
        print(f"{hasattr(PartialGenericModel ,'__args__')=}")

        PartiallySpecializedGenericModel = PartialGenericModel[int, U]
        print(f"{PartiallySpecializedGenericModel.__datamodel_fields__=}")
        print(f"{PartiallySpecializedGenericModel.__parameters__=}")
        print(f"{PartiallySpecializedGenericModel.__args__=}")

        PartiallySpecializedGenericModel(value=[])
        PartiallySpecializedGenericModel(value=[(1, 2)])
        PartiallySpecializedGenericModel(value=[(1, "value")])
        PartiallySpecializedGenericModel(value=[(1, (11, 12))])
        with pytest.raises(TypeError, match=".value'"):
            PartiallySpecializedGenericModel(value=1)
        with pytest.raises(TypeError, match=".value'"):
            PartiallySpecializedGenericModel(value=(1, 2))
        with pytest.raises(TypeError, match=".value'"):
            PartiallySpecializedGenericModel(value=[1.0])
        with pytest.raises(TypeError, match=".value'"):
            PartiallySpecializedGenericModel(value=["1"])

        # TODO(egparedes): after fixing partial nested datamodel specialization
        # noqa: e800 FullySpecializedGenericModel = PartiallySpecializedGenericModel[str]
        # noqa: e800 print(f"{FullySpecializedGenericModel.__datamodel_fields__=}")
        # noqa: e800 print(f"{FullySpecializedGenericModel.__parameters__=}")
        # noqa: e800 print(f"{FullySpecializedGenericModel.__args__=}")

        # noqa: e800 FullySpecializedGenericModel(value=[])
        # noqa: e800 FullySpecializedGenericModel(value=[(1, "value")])
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=1)
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=(1, 2))
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=[1.0])
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=["1"])
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=1)
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=[(1, 2)])
        # noqa: e800 with pytest.raises(TypeError, match=".value'"):
        # noqa: e800     FullySpecializedGenericModel(value=[(1, (11, 12))])

    # Reuse sample_type_data from test_field_type_hint
    @pytest.mark.parametrize(["type_hint", "valid_values", "wrong_values"], SAMPLE_TYPE_DATA)
    def test_concrete_field_type(
        self, type_hint: str, valid_values: Sequence[Any], wrong_values: Sequence[Any]
    ):
        concrete_type: Type = eval(type_hint)
        Model: Type[datamodels.DataModel] = GenericModel[concrete_type]  # type: ignore[valid-type,misc,assignment]  # concrete_type

        for value in valid_values:
            Model(value=value)

        for value in wrong_values:
            with pytest.raises((TypeError, ValueError), match=".value'"):
                Model(value=value)


# ---- Validators ----
def test_non_empty():
    @datamodels.datamodel
    class Model:
        str_value: str = datamodels.field(validator=datamodels.validators.non_empty())
        list_value: List[int] = datamodels.field(validator=datamodels.validators.non_empty())

    Model("AAAA", [1])

    with pytest.raises(ValueError, match="Empty"):
        Model("", [])

    with pytest.raises(ValueError, match="Empty"):
        Model("", [1])

    with pytest.raises(ValueError, match="Empty"):
        Model("AAAA", [])



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Eve framework with general utils for development of DSL toolchains in Python.

The internal dependencies between modules are the following (each module depends
on some of the previous ones):

  0. extended_typing
  1. exceptions, pattern_matching, type_definitions
  2. utils
  3. type_validation
  4. datamodels
  5. trees
  6. concepts
  7. visitors
  8. traits
  9. codegen

"""

from __future__ import annotations  # isort:skip


from .concepts import (
    AnnexManager,
    AnySourceLocation,
    FrozenNode,
    GenericNode,
    Node,
    RootNode,
    SourceLocation,
    SourceLocationGroup,
    SymbolName,
    SymbolRef,
    VType,
    register_annex_user,
)
from .datamodels import (
    Coerced,
    DataModel,
    FrozenModel,
    GenericDataModel,
    Unchecked,
    concretize,
    datamodel,
    field,
    frozenmodel,
)
from .traits import SymbolTableTrait, ValidatedSymbolTableTrait, VisitorWithSymbolTableTrait
from .trees import (
    bfs_walk_items,
    bfs_walk_values,
    post_walk_items,
    post_walk_values,
    pre_walk_items,
    pre_walk_values,
    walk_items,
    walk_values,
)
from .type_definitions import NOTHING, ConstrainedStr, Enum, IntEnum, NothingType, StrEnum
from .visitors import NodeTranslator, NodeVisitor


__all__ = [
    # version
    "__version__",
    "__version_info__",
    # concepts
    "AnnexManager",
    "AnySourceLocation",
    "FrozenNode",
    "RootNode",
    "GenericNode",
    "Node",
    "SourceLocation",
    "SourceLocationGroup",
    "SymbolName",
    "SymbolRef",
    "VType",
    "register_annex_user",
    "# datamodels" "Coerced",
    "DataModel",
    "FrozenModel",
    "GenericDataModel",
    "Unchecked",
    "concretize",
    "datamodel",
    "field",
    "frozenmodel",
    # datamodels
    "Coerced",
    "DataModel",
    "FrozenModel",
    "GenericDataModel",
    "Unchecked",
    "concretize",
    "datamodel",
    "field",
    "frozenmodel",
    # traits
    "SymbolTableTrait",
    "ValidatedSymbolTableTrait",
    "VisitorWithSymbolTableTrait",
    # trees
    "bfs_walk_items",
    "bfs_walk_values",
    "post_walk_items",
    "post_walk_values",
    "pre_walk_items",
    "pre_walk_values",
    "walk_items",
    "walk_values",
    "# type_definition",
    "NOTHING",
    "ConstrainedStr",
    "Enum",
    "IntEnum",
    "NothingType",
    "StrEnum",
    # visitors
    "NodeTranslator",
    "NodeVisitor",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
import copy
import hashlib
import os
import pathlib
import time
import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Protocol, Tuple, Type, Union

from gt4py import storage as gt_storage
from gt4py.cartesian import definitions as gt_definitions
from gt4py.cartesian import utils as gt_utils

from . import pyext_builder
from .module_generator import BaseModuleGenerator, ModuleData, make_args_data_from_gtir


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_builder import StencilBuilder
    from gt4py.cartesian.stencil_object import StencilObject

REGISTRY = gt_utils.Registry()


def from_name(name: str) -> Optional[Type["Backend"]]:
    return REGISTRY.get(name, None)


def register(backend_cls: Type["Backend"]) -> None:
    assert issubclass(backend_cls, Backend) and backend_cls.name is not None

    if isinstance(backend_cls.name, str):
        gt_storage.register(backend_cls.name, backend_cls.storage_info)
        return REGISTRY.register(backend_cls.name, backend_cls)

    else:
        raise ValueError(
            "Invalid 'name' attribute ('{name}') in backend class '{cls}'".format(
                name=backend_cls.name, cls=backend_cls
            )
        )


class Backend(abc.ABC):

    #: Backend name
    name: ClassVar[str]

    #: Backend-specific options:
    #:
    #:  + info:
    #:    - versioning: is versioning on?
    #:    - description [optional]
    #:    - type
    options: ClassVar[Dict[str, Any]]

    #: Backend-specific storage parametrization:
    #:
    #:  - "alignment": in bytes
    #:  - "device": "cpu" | "gpu"
    #:  - "layout_map": callback converting a mask to a layout
    #:  - "is_optimal_layout": callback checking if a storage has compatible layout
    storage_info: ClassVar[gt_storage.layout.LayoutInfo]

    #: Language support:
    #:
    #:  - "computation": name of the language in which the computation is implemented.
    #:  - "bindings": names of supported language bindings / wrappers.
    #:    If a high-level language is compatible out-of-the-box, it should not be listed.
    #:
    #:  Languages should be spelled using the official spelling
    #:  but lower case ("python", "fortran", "rust").
    languages: ClassVar[Optional[Dict[str, Any]]] = None

    # __impl_opts:
    #   "disable-code-generation": bool
    #   "disable-cache-validation": bool

    builder: "StencilBuilder"

    def __init__(self, builder: "StencilBuilder"):
        self.builder = builder

    @classmethod
    def filter_options_for_id(
        cls, options: gt_definitions.BuildOptions
    ) -> gt_definitions.BuildOptions:
        """Filter a copy for options that should not be part of the stencil ID."""
        filtered_options = copy.deepcopy(options)
        id_names = set(name for name, info in cls.options.items() if info["versioning"])
        non_id_names = set(filtered_options.backend_opts.keys()) - id_names
        for name in non_id_names:
            del filtered_options.backend_opts[name]

        return filtered_options

    @abc.abstractmethod
    def load(self) -> Optional[Type["StencilObject"]]:
        """
        Load the stencil class from the generated python module.

        type:
            The generated stencil class after loading through python's import API


        This assumes that :py:meth:`Backend.generate` has been called already on the same stencil.
        The python module therefore already exists and any python extensions have been compiled.
        In case the stencil changed since the last generate call, it will not be rebuilt.
        """
        pass

    @abc.abstractmethod
    def generate(self) -> Type["StencilObject"]:
        """
        Generate the stencil class from GTScript's internal representation.

        Returns
        -------
        type:
            The generated stencil class after loading through python's import API

        In case the stencil class for the given ID is found in cache, this method can avoid
        rebuilding it. Rebuilding can however be triggered through the :code:`options` parameter
        if supported by the backend.
        """
        pass

    @property
    def extra_cache_info(self) -> Dict[str, Any]:
        """Provide additional data to be stored in cache info file (sublass hook)."""
        return {}

    @property
    def extra_cache_validation_keys(self) -> List[str]:
        """List keys from extra_cache_info to be validated during consistency check."""
        return []


class CLIBackendMixin(Backend):
    @abc.abstractmethod
    def generate_computation(self) -> Dict[str, Union[str, Dict]]:
        """
        Generate the computation source code in a way agnostic of the way it is going to be used.

        Returns
        -------
        Dict[str, str | Dict] of source file names / directories -> contents:
            If a key's value is a string it is interpreted as a file name and the value as the
            source code of that file
            If a key's value is a Dict, it is interpreted as a directory name and it's
            value as a nested file hierarchy to which the same rules are applied recursively.
            The root path is relative to the build directory.

        Raises
        ------
        NotImplementedError
            If the backend does not support usage outside of JIT compilation / generation.

        Example
        -------
        .. code-block:: python

            def mystencil(...):
                ...

            options = BuildOptions(name="mystencil", ...)
            ir = frontend.generate(mystencil, {}, options)
            stencil_src = backend.generate_computation(ir, options)

            print(stencil_src)

            # this might be output from a fictional backend:
            {
                "mystencil_project": {
                    "src": {
                        "stencil.cpp",
                        "helpers.cpp"
                    },
                    "include": {
                        "stencil.hpp"
                    },
                }
            }

        This can now be automatically be turned into a folder hierarchy that makes sense
        and can be incorporated into an external build system.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def generate_bindings(self, language_name: str) -> Dict[str, Union[str, Dict]]:
        """
        Generate bindings source code from ``language_name`` to the target language of the backend.

        Returns
        -------
        Analog to :py:meth:`generate_computation` but containing bindings source code, The
        dictionary contains a tree of directories with leaves being a mapping from filename to
        source code pairs, relative to the build directory.

        Raises
        ------
        RuntimeError
            If the backend does not support the bindings language

        """
        languages = getattr(self, "languages", {"bindings": {}})
        name = getattr(self, "name", "")
        if language_name not in languages["bindings"]:
            raise NotImplementedError(
                f"Backend {name} does not implement bindings for {language_name}"
            )
        return {}


class BaseBackend(Backend):

    MODULE_GENERATOR_CLASS: ClassVar[Type["BaseModuleGenerator"]]

    def load(self) -> Optional[Type["StencilObject"]]:
        build_info = self.builder.options.build_info
        if build_info is not None:
            start_time = time.perf_counter()

        stencil_class = None
        if self.builder.stencil_id is not None:
            self.check_options(self.builder.options)
            validate_hash = not self.builder.options._impl_opts.get(
                "disable-cache-validation", False
            )
            if self.builder.caching.is_cache_info_available_and_consistent(
                validate_hash=validate_hash
            ):
                stencil_class = self._load()

        if build_info is not None:
            build_info["load_time"] = time.perf_counter() - start_time

        return stencil_class

    def generate(self) -> Type["StencilObject"]:
        self.check_options(self.builder.options)
        return self.make_module()

    def _load(self) -> Type["StencilObject"]:
        stencil_class_name = self.builder.class_name
        file_name = str(self.builder.module_path)
        stencil_module = gt_utils.make_module_from_file(stencil_class_name, file_name)
        stencil_class = getattr(stencil_module, stencil_class_name)
        stencil_class.__module__ = self.builder.module_qualname
        stencil_class._gt_id_ = self.builder.stencil_id.version
        stencil_class._file_name = file_name
        stencil_class.definition_func = staticmethod(self.builder.definition)

        return stencil_class

    def check_options(self, options: gt_definitions.BuildOptions) -> None:
        assert self.options is not None
        unknown_options = set(options.backend_opts.keys()) - set(self.options.keys())
        if unknown_options:
            warnings.warn(
                f"Unknown options '{unknown_options}' for backend '{self.name}'", RuntimeWarning
            )

    def make_module(
        self,
        **kwargs: Any,
    ) -> Type["StencilObject"]:
        build_info = self.builder.options.build_info
        if build_info is not None:
            start_time = time.perf_counter()

        file_path = self.builder.module_path
        module_source = self.make_module_source(**kwargs)

        if not self.builder.options._impl_opts.get("disable-code-generation", False):
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(module_source)
            self.builder.caching.update_cache_info()

        module = self._load()

        if build_info is not None:
            build_info["module_time"] = time.perf_counter() - start_time

        return module

    def make_module_source(self, *, args_data: Optional[ModuleData] = None, **kwargs: Any) -> str:
        """Generate the module source code with or without stencil id."""
        args_data = args_data or make_args_data_from_gtir(self.builder.gtir_pipeline)
        source = self.MODULE_GENERATOR_CLASS()(args_data, self.builder, **kwargs)
        return source


class MakeModuleSourceCallable(Protocol):
    def __call__(self, *, args_data: Optional[ModuleData] = None, **kwargs: Any) -> str:
        ...


class PurePythonBackendCLIMixin(CLIBackendMixin):
    """Mixin for CLI support for backends deriving from BaseBackend."""

    builder: "StencilBuilder"

    #: stencil python source generator method:
    #:  In order to use this mixin, the backend class must implement
    #:  a :py:meth:`make_module_source` method or derive from
    #:  :py:meth:`BaseBackend`.
    make_module_source: MakeModuleSourceCallable

    def generate_computation(self) -> Dict[str, Union[str, Dict]]:
        file_name = self.builder.module_path.name
        source = self.make_module_source(ir=self.builder.gtir)
        return {str(file_name): source}

    def generate_bindings(self, language_name: str) -> Dict[str, Union[str, Dict]]:
        """Pure python backends typically will not support bindings."""
        return super().generate_bindings(language_name)


class BasePyExtBackend(BaseBackend):
    @property
    def pyext_module_name(self) -> str:
        return self.builder.module_name + "_pyext"

    @property
    def pyext_module_path(self) -> str:
        return self.builder.module_qualname + "_pyext"

    @property
    def pyext_class_name(self) -> str:
        return self.builder.class_name + "_pyext"

    @property
    def pyext_build_dir_path(self) -> pathlib.Path:
        return self.builder.pkg_path.joinpath(self.pyext_module_name + "_BUILD")

    @property
    def extra_cache_info(self) -> Dict[str, Any]:
        pyext_file_path = self.builder.backend_data.get("pyext_file_path", None)
        pyext_md5 = ""
        if pyext_file_path:
            pyext_md5 = hashlib.md5(pathlib.Path(pyext_file_path).read_bytes()).hexdigest()
        return {
            **super().extra_cache_info,
            "pyext_file_path": pyext_file_path,
            "pyext_md5": pyext_md5,
        }

    @property
    def extra_cache_validation_keys(self) -> List[str]:
        keys = super().extra_cache_validation_keys
        if self.extra_cache_info["pyext_md5"]:
            keys.append("pyext_md5")
        return keys

    @abc.abstractmethod
    def generate(self) -> Type["StencilObject"]:
        pass

    def build_extension_module(
        self,
        pyext_sources: Dict[str, Any],
        pyext_build_opts: Dict[str, str],
        *,
        uses_cuda: bool = False,
    ) -> Tuple[str, str]:
        # Build extension module
        pyext_build_path = pathlib.Path(
            os.path.relpath(self.pyext_build_dir_path, pathlib.Path.cwd())
        )
        pyext_build_path.mkdir(parents=True, exist_ok=True)
        sources = []
        for key, source in pyext_sources.items():
            src_file_path = pyext_build_path / key
            src_ext = src_file_path.suffix
            if src_ext not in [".h", ".hpp"]:
                sources.append(str(src_file_path))

            if source is not gt_utils.NOTHING:
                src_file_path.write_text(source)

        pyext_target_file_path = self.builder.pkg_path
        qualified_pyext_name = self.pyext_module_path

        pyext_build_args: Dict[str, Any] = dict(
            name=qualified_pyext_name,
            sources=sources,
            build_path=str(pyext_build_path),
            target_path=str(pyext_target_file_path),
            **pyext_build_opts,
        )

        if uses_cuda:
            module_name, file_path = pyext_builder.build_pybind_cuda_ext(**pyext_build_args)
        else:
            module_name, file_path = pyext_builder.build_pybind_ext(**pyext_build_args)

        assert module_name == qualified_pyext_name

        self.builder.with_backend_data(
            {"pyext_module_name": module_name, "pyext_file_path": file_path}
        )

        return module_name, file_path



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable, Dict, Optional, Sequence, Tuple

from gt4py.cartesian.definitions import StencilID
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.gtc.passes.gtir_definitive_assignment_analysis import (
    check as check_assignments,
)
from gt4py.cartesian.gtc.passes.gtir_dtype_resolver import resolve_dtype
from gt4py.cartesian.gtc.passes.gtir_prune_unused_parameters import prune_unused_parameters
from gt4py.cartesian.gtc.passes.gtir_upcaster import upcast


PASS_T = Callable[[gtir.Stencil], gtir.Stencil]


class GtirPipeline:
    """
    GTIR passes pipeline runs passes in order and allows skipping.

    May only call existing passes and may not contain any pass logic itself.
    """

    _cache: Dict[Tuple[StencilID, Tuple[PASS_T, ...]], gtir.Stencil] = {}

    def __init__(self, node: gtir.Stencil, stencil_id: StencilID):
        self.gtir = node
        self._stencil_id = stencil_id

    @property
    def stencil_id(self) -> StencilID:
        return self._stencil_id

    def steps(self) -> Sequence[PASS_T]:
        return [check_assignments, prune_unused_parameters, resolve_dtype, upcast]

    def apply(self, steps: Sequence[PASS_T]) -> gtir.Stencil:
        result = self.gtir
        for step in steps:
            result = step(result)
        return result

    def _get_cached(self, steps: Sequence[PASS_T]) -> Optional[gtir.Stencil]:
        return self._cache.get((self.stencil_id, tuple(steps)))

    def _set_cached(self, steps: Sequence[PASS_T], node: gtir.Stencil) -> gtir.Stencil:
        return self._cache.setdefault((self.stencil_id, tuple(steps)), node)

    def full(self, skip: Sequence[PASS_T] = None) -> gtir.Stencil:
        skip = skip or []
        pipeline = [step for step in self.steps() if step not in skip]
        return self._get_cached(pipeline) or self._set_cached(pipeline, self.apply(pipeline))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import collections
import enum
import numbers
import operator

from gt4py.cartesian.gtc.utils import filter_mask, interpolate_mask


class CartesianSpace:
    @enum.unique
    class Axis(enum.Enum):
        I = 0  # noqa: E741  # Do not use variables named 'I', 'O', or 'l'
        J = 1
        K = 2

        def __str__(self):
            return self.name

    names = [ax.name for ax in Axis]
    ndim = len(names)


class NumericTuple(tuple):
    """N-dimensional like vector implemented as a subclass of the tuple builtin."""

    __slots__ = ()

    _DEFAULT = 0

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(isinstance(d, numbers.Number) for d in value)
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])

    @classmethod
    def is_valid(cls, value, *, ndims=(1, None)):
        if isinstance(ndims, numbers.Integral):
            ndims = tuple([ndims] * 2)
        elif not isinstance(ndims, tuple) or len(ndims) != 2:
            raise ValueError("Invalid 'ndims' definition ({})".format(ndims))

        try:
            cls._check_value(value, ndims)
        except Exception:
            return False
        else:
            return True

    @classmethod
    def zeros(cls, ndims=CartesianSpace.ndim):
        return cls([0] * ndims, ndims=(ndims, ndims))

    @classmethod
    def ones(cls, ndims=CartesianSpace.ndim):
        return cls([1] * ndims, ndims=(ndims, ndims))

    @classmethod
    def from_k(cls, value, ndims=CartesianSpace.ndim):
        return cls([value] * ndims, ndims=(ndims, ndims))

    @classmethod
    def from_mask(cls, seq, mask, default=None):
        if default is None:
            default = cls._DEFAULT
        return cls(interpolate_mask(seq, mask, default))

    @classmethod
    def from_value(cls, value):
        if isinstance(value, collections.abc.Iterable):
            return cls(list(value))
        else:
            return cls.from_k(value)

    def __new__(cls, sizes, *args, ndims=None):
        if len(args) > 0:
            sizes = [sizes, *args]

        if ndims is None:
            ndims = tuple([len(sizes)] * 2)
        elif isinstance(ndims, int):
            ndims = tuple([ndims] * 2)
        elif not isinstance(ndims, tuple) or len(ndims) != 2:
            raise ValueError("Invalid 'ndims' definition ({})".format(ndims))

        try:
            cls._check_value(sizes, ndims=ndims)
        except Exception as e:
            raise TypeError("Invalid {} definition".format(cls.__name__)) from e
        else:
            return super().__new__(cls, sizes)

    def __getattr__(self, name):
        try:
            value = self[CartesianSpace.Axis.symbols.index(name)]
        except (IndexError, ValueError) as e:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(self.__class__.__name__, name)
            ) from e
        else:
            return value

    def __getitem__(self, item):
        if isinstance(item, CartesianSpace.Axis):
            item = item.value
        return tuple.__getitem__(self, item)

    def __add__(self, other):
        """Element-wise addition."""
        return self._apply(self._broadcast(other), operator.add)

    def __sub__(self, other):
        """Element-wise substraction."""
        return self._apply(self._broadcast(other), operator.sub)

    def __mul__(self, other):
        """Element-wise multiplication."""
        return self._apply(self._broadcast(other), operator.mul)

    def __floordiv__(self, other):
        """Element-wise integer division."""
        return self._apply(self._broadcast(other), operator.floordiv)

    def __and__(self, other):
        """Element-wise intersection operation."""
        return self._apply(other, min)

    def __or__(self, other):
        """Element-wise union operation."""
        return self._apply(other, max)

    def __lt__(self, other):
        """No element can be greater, but if any element is smaller, return True."""
        return self._compare(
            self._broadcast(other),
            lambda a, b: -1 if a < b else (1 if a > b else 0),
            lambda items: any(i < 0 for i in items) and not any(i > 0 for i in items),
        )

    def __le__(self, other):
        """Element-wise comparison."""
        return self._compare(
            self._broadcast(other),
            lambda a, b: -1 if a < b else (1 if a > b else 0),
            lambda items: all(i <= 0 for i in items),
        )

    def __eq__(self, other):
        """Element-wise comparison."""
        return self._compare(self._broadcast(other), operator.eq, all)

    def __ne__(self, other):
        """Element-wise comparison."""
        return not self._compare(self._broadcast(other), operator.eq, all)

    def __gt__(self, other):
        """No element can be smaller, but if any element is larger, return True."""
        return self._compare(
            self._broadcast(other),
            lambda a, b: 1 if a > b else (-1 if a < b else 0),
            lambda items: any(i > 0 for i in items) and not any(i < 0 for i in items),
        )

    def __ge__(self, other):
        """Element-wise comparison."""
        return self._compare(
            self._broadcast(other),
            lambda a, b: 1 if a > b else (-1 if a < b else 0),
            lambda items: all(i >= 0 for i in items),
        )

    def __repr__(self):
        return "{cls_name}({value})".format(
            cls_name=type(self).__name__, value=tuple.__repr__(self)
        )

    def __hash__(self):
        return tuple.__hash__(self)

    def __str__(self):
        return tuple.__repr__(self)

    @property
    def __dict__(self):
        """Ordered mapping from axes names to their values."""
        return collections.OrderedDict(zip(CartesianSpace.Axis.symbols, self))

    @property
    def ndims(self):
        return len(self)

    def union(self, other):
        if not isinstance(other, type(self)):
            other = type(self)(other)
        return self.__or__(other)

    def intersection(self, other):
        if not isinstance(other, type(self)):
            other = type(self)(other)
        return self.__and__(other)

    def _apply(self, other, func):
        if not isinstance(other, type(self)) or len(self) != len(other):
            raise ValueError("Incompatible instance '{obj}'".format(obj=other))

        return type(self)([func(a, b) for a, b in zip(self, other)])

    def _broadcast(self, value):
        if isinstance(value, int):
            value = type(self)([value] * self.ndims)
        elif type(self).is_valid(value, ndims=self.ndims):
            value = type(self)(value)

        return value

    def _compare(self, other, op, reduction_op):
        if len(self) != len(other):  # or not isinstance(other, type(self))
            raise ValueError("Incompatible instance '{obj}'".format(obj=other))

        return reduction_op(op(a, b) for a, b in zip(self, other))

    def filter_mask(self, mask):
        return type(self)(filter_mask(self, mask))


class Index(NumericTuple):
    """Index in a grid (all elements are ints)."""

    __slots__ = ()

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(isinstance(d, numbers.Integral) for d in value)
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])


class Shape(NumericTuple):
    """Shape of a n-dimensional grid (all elements are int >= 0)."""

    __slots__ = ()

    _DEFAULT = 1

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(isinstance(d, numbers.Integral) and d >= 0 for d in value)
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])


class FrameTuple(tuple):
    """N-dimensional list of pairs of numbers representing offsets around one central origin."""

    __slots__ = ()

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(
            len(r) == 2 and isinstance(r[0], numbers.Number) and isinstance(r[1], numbers.Number)
            for r in value
        )
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])

    @classmethod
    def is_valid(cls, value, *, ndims=(1, None)):
        if isinstance(ndims, int):
            ndims = tuple([ndims] * 2)
        elif not isinstance(ndims, tuple) or len(ndims) != 2:
            raise ValueError("Invalid 'ndims' definition ({})".format(ndims))

        try:
            cls._check_value(value, ndims)
        except Exception:
            return False
        else:
            return True

    @classmethod
    def zeros(cls, ndims=CartesianSpace.ndim):
        return cls([(0, 0)] * ndims, ndims=(ndims, ndims))

    @classmethod
    def ones(cls, ndims=CartesianSpace.ndim):
        return cls([(1, 1)] * ndims, ndims=(ndims, ndims))

    @classmethod
    def from_k(cls, value_pair, ndims=CartesianSpace.ndim):
        return cls([value_pair] * ndims, ndims=(ndims, ndims))

    @classmethod
    def from_lower_upper(cls, lower, upper):
        ndims = max(len(lower), len(upper))
        return cls([(lower[i], upper[i]) for i in range(ndims)], ndims=(ndims, ndims))

    def __new__(cls, ranges, *args, ndims=None):
        if len(args) > 0:
            ranges = [ranges, *args]

        if ndims is None:
            ndims = tuple([len(ranges)] * 2)
        try:
            cls._check_value(ranges, ndims=ndims)
        except Exception as e:
            raise TypeError("Invalid definition") from e
        else:
            return super().__new__(cls, ranges)

    def __getattr__(self, name):
        try:
            value = self[CartesianSpace.Axis.symbols.index(name)]
        except (IndexError, ValueError) as e:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(self.__class__.__name__, name)
            ) from e
        else:
            return value

    def __getitem__(self, item):
        if isinstance(item, CartesianSpace.Axis):
            item = item.value
        return tuple.__getitem__(self, item)

    def __add__(self, other):
        """Element-wise addition."""
        return self._apply(self._broadcast(other), lambda a, b: a + b)

    def __sub__(self, other):
        """Element-wise substraction."""
        return self._apply(self._broadcast(other), lambda a, b: a - b)

    def __and__(self, other):
        """Element-wise intersection operation."""
        return self._apply(other, min, min)

    def __or__(self, other):
        """Element-wise union operation."""
        return self._apply(other, max, max)

    def __lt__(self, other):
        """Element-wise comparison."""
        return self._compare(self._broadcast(other), operator.lt)

    def __le__(self, other):
        """Element-wise comparison."""
        return self._compare(self._broadcast(other), operator.le)

    def __eq__(self, other):
        """Element-wise comparison."""
        return self._compare(self._broadcast(other), operator.eq)

    def __ne__(self, other):
        """Element-wise comparison."""
        return not self._compare(self._broadcast(other), operator.eq)

    def __gt__(self, other):
        """Element-wise comparison."""
        return self._compare(self._broadcast(other), operator.gt)

    def __ge__(self, other):
        """Element-wise comparison."""
        return self._compare(self._broadcast(other), operator.ge)

    def __repr__(self):
        return "{cls_name}({value})".format(
            cls_name=self.__class__.__name__, value=tuple.__repr__(self)
        )

    def __hash__(self):
        return tuple.__hash__(self)

    def __str__(self):
        return tuple.__repr__(self)

    @property
    def __dict__(self):
        """Ordered mapping from axes names to their values."""
        return collections.OrderedDict(zip(CartesianSpace.Axis.symbols, self))

    @property
    def ndims(self):
        return len(self)

    @property
    def is_symmetric(self):
        return all(d[0] == d[1] for d in self)

    @property
    def is_zero(self):
        return all(d[0] == d[1] == 0 for d in self)

    @property
    def lower_indices(self):
        return NumericTuple(*(d[0] for d in self))

    @property
    def upper_indices(self):
        return NumericTuple(*(d[1] for d in self))

    def append(self, point):
        other = self.__class__([(i, i) for i in point])
        return self.__or__(other)

    def concatenate(self, other):
        if not isinstance(other, FrameTuple):
            other = type(self)(other)
        return self.__add__(other)

    def union(self, other):
        if not isinstance(other, FrameTuple):
            other = type(self)(other)
        return self.__or__(other)

    def intersection(self, other):
        if not isinstance(other, FrameTuple):
            other = type(self)(other)
        return self.__and__(other)

    def _apply(self, other, left_func, right_func=None):
        if not isinstance(other, FrameTuple) or len(self) != len(other):
            raise ValueError("Incompatible instance '{obj}'".format(obj=other))

        right_func = right_func or left_func
        return type(self)(
            [tuple([left_func(a[0], b[0]), right_func(a[1], b[1])]) for a, b in zip(self, other)]
        )

    def _reduce(self, reduce_func, out_type=tuple):
        return out_type([reduce_func(d[0], d[1]) for d in self])

    def _compare(self, other, left_op, right_op=None):
        if len(self) != len(other):  # or not isinstance(other, Frame)
            raise ValueError("Incompatible instance '{obj}'".format(obj=other))

        right_op = right_op or left_op
        return all(left_op(a[0], b[0]) and right_op(a[1], b[1]) for a, b in zip(self, other))

    def _broadcast(self, value):
        if isinstance(value, int):
            value = type(self)([(value, value)] * self.ndims)
        elif type(self).is_valid(value, ndims=self.ndims):
            value = type(self)(value)

        return value


class Boundary(FrameTuple):
    """Frame size around one central origin (pairs of integers).

    Negative numbers represent a boundary region substracting from
    the wrapped area.
    """

    __slots__ = ()

    def __hash__(self):
        return tuple.__hash__(self)

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(
            len(r) == 2 and isinstance(r[0], int) and isinstance(r[1], int)
            # and r[0] >= 0 and r[1] >= 0
            for r in value
        )
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])

    @classmethod
    def from_offset(cls, offset):
        if not Index.is_valid(offset):
            raise ValueError("Invalid offset value ({})".format(offset))
        return cls([(-1 * min(0, i), max(0, i)) for i in offset])

    @property
    def frame_size(self):
        """Sizes of the boundary area, i.e., excluding the central element of the frame."""
        return Shape(tuple(d[1] + d[0] for d in self))

    @property
    def shape(self):
        """Total shape of the frame area."""
        return Shape(tuple(d[1] + d[0] + 1 for d in self))


class Extent(FrameTuple):
    """Region defined by the smallest and the largest offsets.

    Size of boundary regions are expressed as minimum and maximum relative offsets related
    to the computation position. For example the frame for a stencil accessing 3 elements
    at the left of computation point and two at the right in the X axis would be: (-3, 2).
    If the stencil only has accesses with negative indices, the upper boundary would be then
    a negative value.
    """

    __slots__ = ()

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(
            len(r) == 2
            and isinstance(r[0], (int, type(None)))
            and isinstance(r[1], (int, type(None)))
            and (r[0] is None or r[1] is None or r[1] <= r[1])
            for r in value
        )
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])

    @classmethod
    def empty(cls, ndims=CartesianSpace.ndim):
        return cls([(None, None)] * ndims)

    @classmethod
    def from_offset(cls, offset):
        if not Index.is_valid(offset):
            raise ValueError("Invalid offset value ({})".format(offset))
        return cls([(i, i) for i in offset])

    def __and__(self, other):
        """Intersection operation."""
        return self._apply(other, max, min)

    def __or__(self, other):
        """Union operation."""
        return self._apply(other, min, max)

    def __lt__(self, other):
        return self._compare(other, operator.gt, operator.lt)

    def __le__(self, other):
        return self._compare(other, operator.ge, operator.le)

    def __eq__(self, other):
        return self._compare(other, operator.eq)

    def __ne__(self, other):
        return not self._compare(other, operator.eq)

    def __gt__(self, other):
        return self._compare(other, operator.lt, operator.gt)

    def __ge__(self, other):
        return self._compare(other, operator.le, operator.ge)

    def __hash__(self):
        return tuple.__hash__(self)

    @property
    def frame_size(self):
        """Sizes of the boundary area, i.e., excluding the central element of the frame."""
        return Shape(tuple(d[1] - d[0] for d in self))

    @property
    def shape(self):
        """Total shape of the frame area."""
        return Shape(tuple(d[1] - d[0] + 1 for d in self))

    def to_boundary(self):
        return Boundary([(-1 * d[0], d[1]) for d in self])

    def _apply(self, other, left_func, right_func=None):
        if not isinstance(other, FrameTuple) or len(self) != len(other):
            raise ValueError("Incompatible instance '{obj}'".format(obj=other))

        right_func = right_func or left_func
        result = [None] * len(self)
        for i, (a, b) in enumerate(zip(self, other)):
            if a[0] is None:
                left = b[0]
            elif b[0] is None:
                left = a[0]
            else:
                left = left_func(a[0], b[0])

            if a[1] is None:
                right = b[1]
            elif b[0] is None:
                right = a[1]
            else:
                right = right_func(a[1], b[1])

            result[i] = (left, right)

        return type(self)(result)


class CenteredExtent(Extent):
    """Region defined by the largest negative offset (or zero) and the largest positive offset (or zero).

    Size of boundary regions are expressed as minimum and maximum relative offsets related
    to the computation position. For example the frame for a stencil accessing 3 elements
    at the left of computation point and two at the right in the X axis would be: (-3, 2).
    If the stencil only has accesses with negative indices, the upper boundary would be 0.
    """

    __slots__ = ()

    @classmethod
    def _check_value(cls, value, ndims):
        assert isinstance(value, collections.abc.Collection), "Invalid collection"
        assert all(
            len(r) == 2 and isinstance(r[0], int) and isinstance(r[1], int) and r[0] <= 0 <= r[1]
            for r in value
        )
        assert ndims[0] <= len(value) and (ndims[1] is None or len(value) <= ndims[1])

    @classmethod
    def empty(cls, ndims=CartesianSpace.ndim):
        return cls.zeros(ndims)

    @classmethod
    def from_offset(cls, offset):
        if not Index.is_valid(offset):
            raise ValueError("Invalid offset value ({})".format(offset))
        return cls([(min(i, 0), max(i, 0)) for i in offset])

    def to_boundary(self):
        return Boundary([(-1 * min(d[0], 0), max(d[1], 0)) for d in self])



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import collections
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Set, Tuple

from gt4py import eve
from gt4py.cartesian.gtc import common, oir

from .utils import AccessCollector, symbol_name_creator


"""Utilities for cache detection and modifications on the OIR level.

Caches in GridTools terminology (with roots in STELLA) are local replacements
of temporary fields, which are usually stored as full fields in global memory,
by smaller buffers with more reuse (and faster memory on GPUs) or by registers.

IJ-caches are used in purely horizontal k-parallel stencils and replace the
fully 3D temporary field by 2D buffers (in block-local shared memory on GPUs
and thread-local memory on CPUs).

K-caches are used in vertical sweeps, most notably tridiagonal solvers. They
cache values from vertical accesses in a ring buffer, that is shifted on each
loop iteration. K-cache values can optionally be filled with values from global
memory or flushed to global memory, then acting like normal temporary fields,
but only requiring one read operation per loop iteration, independent of the
number of vertical accesses. In cases, where a field is always written before
read, the fill is not required. Similar, if a field is not read later, the
flushes can be omitted. When fills and flushes are omitted, all field accesses
are replaced by just a few registers.

Note that filling and flushing k-caches can always be replaced by a local
(non-filling or flushing) k-cache plus additional filling and flushing
statements.

"""


class IJCacheDetection(eve.NodeTranslator):
    def visit_VerticalLoop(
        self, node: oir.VerticalLoop, *, local_tmps: Set[str], **kwargs: Any
    ) -> oir.VerticalLoop:
        if node.loop_order != common.LoopOrder.PARALLEL or not local_tmps:
            return node

        def already_cached(field: str) -> bool:
            return any(c.name == field for c in node.caches)

        def has_vertical_offset(offsets: Set[Tuple[int, int, int]]) -> bool:
            return any(offset[2] != 0 for offset in offsets)

        accesses = AccessCollector.apply(node).cartesian_accesses().offsets()
        cacheable = {
            field
            for field, offsets in accesses.items()
            if field in local_tmps
            and not already_cached(field)
            and not has_vertical_offset(offsets)
        }
        caches = self.visit(node.caches, **kwargs) + [
            oir.IJCache(name=field) for field in cacheable
        ]
        return oir.VerticalLoop(
            sections=node.sections,
            loop_order=node.loop_order,
            caches=caches,
            loc=node.loc,
        )

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        vertical_loops = node.walk_values().if_isinstance(oir.VerticalLoop)
        counts: collections.Counter = sum(
            (
                collections.Counter(
                    vertical_loop.walk_values()
                    .if_isinstance(oir.FieldAccess)
                    .getattr("name")
                    .if_in({tmp.name for tmp in node.declarations})
                    .to_set()
                )
                for vertical_loop in vertical_loops
            ),
            collections.Counter(),
        )
        local_tmps = {tmp for tmp, count in counts.items() if count == 1}
        return self.generic_visit(node, local_tmps=local_tmps, **kwargs)


@dataclass
class KCacheDetection(eve.NodeTranslator):
    max_cacheable_offset: int = 5

    def visit_VerticalLoop(self, node: oir.VerticalLoop, **kwargs: Any) -> oir.VerticalLoop:
        # k-caches are restricted to loops with a single horizontal region as all regions without
        # horizontal offsets should be merged before anyway and this restriction allows for easier
        # conversion of fill and flush caches to local caches later
        if node.loop_order == common.LoopOrder.PARALLEL or any(
            len(section.horizontal_executions) != 1 for section in node.sections
        ):
            return self.generic_visit(node, **kwargs)

        all_accesses = AccessCollector.apply(node)
        fields_with_variable_reads = {
            field
            for field, offsets in all_accesses.offsets().items()
            if any(off[2] is None for off in offsets)
        }

        def accessed_more_than_once(offsets: Set[Any]) -> bool:
            return len(offsets) > 1

        def already_cached(field: str) -> bool:
            return field in {c.name for c in node.caches}

        # TODO(fthaler): k-caches with non-zero ij offsets?
        def has_horizontal_offset(offsets: Set[Tuple[int, int, int]]) -> bool:
            return any(offset[:2] != (0, 0) for offset in offsets)

        def offsets_within_limits(offsets: Set[Tuple[int, int, int]]) -> bool:
            return all(abs(offset[2]) <= self.max_cacheable_offset for offset in offsets)

        def has_variable_offset_reads(field: str) -> bool:
            return field in fields_with_variable_reads

        accesses = all_accesses.cartesian_accesses().offsets()
        cacheable = {
            field
            for field, offsets in accesses.items()
            if not already_cached(field)
            and not has_variable_offset_reads(field)
            and accessed_more_than_once(offsets)
            and not has_horizontal_offset(offsets)
            and offsets_within_limits(offsets)
        }
        caches = self.visit(node.caches, **kwargs) + [
            oir.KCache(name=field, fill=True, flush=True) for field in cacheable
        ]
        return oir.VerticalLoop(
            loop_order=node.loop_order,
            sections=node.sections,
            caches=caches,
            loc=node.loc,
        )


class PruneKCacheFills(eve.NodeTranslator):
    """Prunes unneeded k-cache fills.

    A fill is classified as required if at least one of the two following conditions holds in any of the loop sections:
    * There is a read with offset in the direction of looping.
    * The first centered access is a read access.
    If none of the conditions holds for any loop section, the fill is considered as unneeded.
    """

    def visit_KCache(self, node: oir.KCache, *, pruneable: Set[str], **kwargs: Any) -> oir.KCache:
        if node.name in pruneable:
            return oir.KCache(name=node.name, fill=False, flush=node.flush)
        return self.generic_visit(node, **kwargs)

    def visit_VerticalLoop(self, node: oir.VerticalLoop, **kwargs: Any) -> oir.VerticalLoop:
        filling_fields = {c.name for c in node.caches if isinstance(c, oir.KCache) and c.fill}
        if not filling_fields:
            return self.generic_visit(node, **kwargs)
        assert node.loop_order != common.LoopOrder.PARALLEL

        def pruneable_fields(section: oir.VerticalLoopSection) -> Set[str]:
            accesses = AccessCollector.apply(section).cartesian_accesses()
            offsets = accesses.offsets()
            center_accesses = [a for a in accesses.ordered_accesses() if a.offset == (0, 0, 0)]

            def requires_fill(field: str) -> bool:
                if field not in offsets:
                    return False
                k_offsets = (o[2] for o in offsets[field])
                if node.loop_order == common.LoopOrder.FORWARD and max(k_offsets) > 0:
                    return True
                if node.loop_order == common.LoopOrder.BACKWARD and min(k_offsets) < 0:
                    return True
                try:
                    return next(a.is_read for a in center_accesses if a.field == field)
                except StopIteration:
                    return False

            return {field for field in filling_fields if not requires_fill(field)}

        pruneable = set.intersection(*(pruneable_fields(section) for section in node.sections))

        # Consider accesses outside of loop interval
        # Note: those accesses would not require a fill in the whole loop interval in in theory
        # but restriction to the GridTools k-cache types makes this necessary
        # E.g. consider the following loop with k = 1, 2:
        # k | no fill required | initial fill required |
        # - | ---------------- | --------------------- |
        # 1 | a[k] = init      | a[k] = a[k - 1]       |
        # 2 | a[k] = a[k - 1]  | a[k] = a[k - 1]       |
        # In the first case, no fill is required. In the second case, a[0] has to be filled on
        # level k = 1. On level k = 2, no fill would be necessary, but GridTools C++ does not
        # support a k-cache type that only fills on the initial level, so fills are inserted on all
        # levels here.

        first_section_offsets = (
            AccessCollector.apply(node.sections[0]).cartesian_accesses().offsets()
        )
        last_section_offsets = (
            AccessCollector.apply(node.sections[-1]).cartesian_accesses().offsets()
        )
        for field in list(pruneable):
            first_k_offsets = (o[2] for o in first_section_offsets.get(field, {(0, 0, 0)}))
            last_k_offsets = (o[2] for o in last_section_offsets.get(field, {(0, 0, 0)}))
            if node.loop_order == common.LoopOrder.BACKWARD:
                first_k_offsets, last_k_offsets = last_k_offsets, first_k_offsets
            if min(first_k_offsets) < 0 or max(last_k_offsets) > 0:
                pruneable.remove(field)

        return self.generic_visit(node, pruneable=pruneable, **kwargs)


class PruneKCacheFlushes(eve.NodeTranslator):
    """Prunes unneeded k-cache flushes.

    A flush is classified as unneeded under the following conditions:
    * All accesses to the field are read-only in the current vertical loop.
    * There are no read accesses to the field in a following loop.
    """

    def visit_KCache(self, node: oir.KCache, *, pruneable: Set[str], **kwargs: Any) -> oir.KCache:
        if node.name in pruneable:
            return oir.KCache(name=node.name, fill=node.fill, flush=False, loc=node.loc)
        return self.generic_visit(node, **kwargs)

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        accesses = [AccessCollector.apply(vertical_loop) for vertical_loop in node.vertical_loops]
        vertical_loops = []
        for i, vertical_loop in enumerate(node.vertical_loops):
            flushing_fields = {
                str(c.name) for c in vertical_loop.caches if isinstance(c, oir.KCache) and c.flush
            }
            read_only_fields = flushing_fields & (
                accesses[i].read_fields() - accesses[i].write_fields()
            )
            future_reads: Set[str] = set()
            future_reads = future_reads.union(*(acc.read_fields() for acc in accesses[i + 1 :]))
            tmps_without_reuse = (
                flushing_fields & {str(d.name) for d in node.declarations}
            ) - future_reads
            pruneable = read_only_fields | tmps_without_reuse
            vertical_loops.append(self.visit(vertical_loop, pruneable=pruneable, **kwargs))
        return oir.Stencil(
            name=node.name,
            params=self.visit(node.params, **kwargs),
            vertical_loops=vertical_loops,
            declarations=node.declarations,
            loc=node.loc,
        )


class FillFlushToLocalKCaches(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    """Converts fill and flush k-caches to local k-caches.

    For each cached field, the following actions are performed:
    1. A new locally-k-cached temporary is introduced.
    2. All accesses to the original field are replaced by accesses to this temporary.
    3. Loop sections are split where necessary to allow single-level loads whereever possible.
    3. Fill statements from the original field to the temporary are introduced.
    4. Flush statements from the temporary to the original field are introduced.
    """

    def visit_FieldAccess(
        self, node: oir.FieldAccess, *, name_map: Dict[str, str], **kwargs: Any
    ) -> oir.FieldAccess:
        if node.name in name_map:
            return oir.FieldAccess(
                name=name_map[node.name],
                data_index=node.data_index,
                dtype=node.dtype,
                offset=node.offset,
                loc=node.loc,
            )
        return node

    def visit_VerticalLoopSection(
        self, node: oir.VerticalLoopSection, **kwargs: Any
    ) -> oir.VerticalLoopSection:
        if len(node.horizontal_executions) > 1:
            raise NotImplementedError("Multiple horizontal_executions are not supported")
        return self.generic_visit(node, **kwargs)

    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        *,
        name_map: Dict[str, str],
        fills: List[oir.Stmt],
        flushes: List[oir.Stmt],
        **kwargs: Any,
    ) -> oir.HorizontalExecution:
        return oir.HorizontalExecution(
            body=fills + self.visit(node.body, name_map=name_map, **kwargs) + flushes,
            declarations=node.declarations,
            loc=node.loc,
        )

    @staticmethod
    def _fill_limits(
        loop_order: common.LoopOrder, section: oir.VerticalLoopSection
    ) -> Dict[str, Tuple[int, int]]:
        """Direction-normalized min and max read accesses for each accessed field.

        Args:
            loop_order: forward or backward order.
            section: loop section to split.

        Returns:
            A dict, mapping field names to min and max read offsets relative to loop order (i.e., positive means in the direction of the loop order).

        """

        def directional_k_offset(offset: Tuple[int, int, int]) -> int:
            """Positive k-offset for forward loops, negative for backward."""
            return offset[2] if loop_order == common.LoopOrder.FORWARD else -offset[2]

        read_offsets = AccessCollector.apply(section).cartesian_accesses().read_offsets()
        return {
            field: (
                min(directional_k_offset(o) for o in offsets),
                max(directional_k_offset(o) for o in offsets),
            )
            for field, offsets in read_offsets.items()
        }

    @staticmethod
    def _requires_splitting(interval: oir.Interval) -> bool:
        """Check if an interval is larger than one level and thus may require splitting."""
        if interval.start.level != interval.end.level:
            return True
        return abs(interval.end.offset - interval.start.offset) > 1

    @staticmethod
    def _split_entry_level(
        loop_order: common.LoopOrder,
        section: oir.VerticalLoopSection,
        new_symbol_name: Callable[[str], str],
    ) -> Tuple[oir.VerticalLoopSection, oir.VerticalLoopSection]:
        """Split the entry level of a loop section.

        Args:
            loop_order: forward or backward order.
            section: loop section to split.

        Returns:
            Two loop sections.
        """
        assert loop_order in (common.LoopOrder.FORWARD, common.LoopOrder.BACKWARD)
        if loop_order == common.LoopOrder.FORWARD:
            bound = common.AxisBound(
                level=section.interval.start.level, offset=section.interval.start.offset + 1
            )
            entry_interval = oir.Interval(start=section.interval.start, end=bound)
            rest_interval = oir.Interval(start=bound, end=section.interval.end)
        else:
            bound = common.AxisBound(
                level=section.interval.end.level, offset=section.interval.end.offset - 1
            )
            entry_interval = oir.Interval(start=bound, end=section.interval.end)
            rest_interval = oir.Interval(start=section.interval.start, end=bound)
        decls = list(section.walk_values().if_isinstance(oir.Decl))
        decls_map = {decl.name: new_symbol_name(decl.name) for decl in decls}

        class FixSymbolNameClashes(eve.NodeTranslator):
            def visit_ScalarAccess(self, node: oir.ScalarAccess) -> oir.ScalarAccess:
                if node.name not in decls_map:
                    return node
                return oir.ScalarAccess(name=decls_map[node.name], dtype=node.dtype)

            def visit_LocalScalar(self, node: oir.LocalScalar) -> oir.LocalScalar:
                return oir.LocalScalar(name=decls_map[node.name], dtype=node.dtype)

        return (
            oir.VerticalLoopSection(
                interval=entry_interval,
                horizontal_executions=FixSymbolNameClashes().visit(section.horizontal_executions),
                loc=section.loc,
            ),
            oir.VerticalLoopSection(
                interval=rest_interval,
                horizontal_executions=section.horizontal_executions,
                loc=section.loc,
            ),
        )

    @classmethod
    def _split_section_with_multiple_fills(
        cls,
        loop_order: common.LoopOrder,
        section: oir.VerticalLoopSection,
        filling_fields: Iterable[str],
        first_unfilled: Dict[str, int],
        new_symbol_name: Callable[[str], str],
    ) -> Tuple[Tuple[oir.VerticalLoopSection, ...], Dict[str, int]]:
        """Split loop sections that require multiple fills.

        Args:
            loop_order: forward or backward order.
            section: loop section to split.
            filling_fields: fields that are using fill caches.
            first_unfilled: direction-normalized offset of the first unfilled cache entry for each field.

        Returns:
            A list of sections and an updated `first_unfilled` map.
        """
        fill_limits = cls._fill_limits(loop_order, section)
        max_required_fills = 0
        for field in filling_fields:
            lmin, lmax = fill_limits.get(field, (0, 0))
            required_fills = lmax + 1 - max(lmin, first_unfilled.get(field, lmin))
            max_required_fills = max(required_fills, max_required_fills)
            first_unfilled[field] = lmax
        if max_required_fills > 1 and cls._requires_splitting(section.interval):
            return cls._split_entry_level(loop_order, section, new_symbol_name), first_unfilled
        return (section,), first_unfilled

    @classmethod
    def _fill_stmts(
        cls,
        loop_order: common.LoopOrder,
        section: oir.VerticalLoopSection,
        filling_fields: Dict[str, str],
        first_unfilled: Dict[str, int],
        symtable: Dict[str, Any],
    ) -> Tuple[List[oir.AssignStmt], Dict[str, int]]:
        """Generate fill statements for the given loop section.

        Args:
            loop_order: forward or backward order.
            section: loop section to split.
            filling_fields: mapping from field names to cache names.
            first_unfilled: direction-normalized offset of the first unfilled cache entry for each field.

        Returns:
            A list of fill statements and an updated `first_unfilled` map.
        """
        fill_limits = cls._fill_limits(loop_order, section)
        fill_stmts = []
        for field, cache in filling_fields.items():
            lmin, lmax = fill_limits.get(field, (0, 0))
            lmin = max(lmin, first_unfilled.get(field, lmin))
            for offset in range(lmin, lmax + 1):
                k_offset = common.CartesianOffset(
                    i=0,
                    j=0,
                    k=offset if loop_order == common.LoopOrder.FORWARD else -offset,
                )
                fill_stmts.append(
                    oir.AssignStmt(
                        left=oir.FieldAccess(
                            name=cache, dtype=symtable[field].dtype, offset=k_offset
                        ),
                        right=oir.FieldAccess(
                            name=field, dtype=symtable[field].dtype, offset=k_offset
                        ),
                    )
                )
            first_unfilled[field] = lmax
        return fill_stmts, first_unfilled

    @classmethod
    def _flush_stmts(
        cls,
        loop_order: common.LoopOrder,
        section: oir.VerticalLoopSection,
        flushing_fields: Dict[str, str],
        symtable: Dict[str, Any],
    ) -> List[oir.AssignStmt]:
        """Generate flush statements for the given loop section.

        Args:
            loop_order: forward or backward order.
            section: loop section to split.
            flushing_fields: mapping from field names to cache names.

        Returns:
            A list of flush statements.
        """
        write_fields = AccessCollector.apply(section).write_fields()
        flush_stmts = []
        for field, cache in flushing_fields.items():
            if field in write_fields:
                flush_stmts.append(
                    oir.AssignStmt(
                        left=oir.FieldAccess(
                            name=field,
                            dtype=symtable[field].dtype,
                            offset=common.CartesianOffset.zero(),
                        ),
                        right=oir.FieldAccess(
                            name=cache,
                            dtype=symtable[field].dtype,
                            offset=common.CartesianOffset.zero(),
                        ),
                    )
                )
        return flush_stmts

    def visit_VerticalLoop(
        self,
        node: oir.VerticalLoop,
        *,
        new_tmps: List[oir.Temporary],
        symtable: Dict[str, Any],
        new_symbol_name: Callable[[str], str],
        **kwargs: Any,
    ) -> oir.VerticalLoop:
        filling_fields: Dict[str, str] = {
            c.name: new_symbol_name(c.name)
            for c in node.caches
            if isinstance(c, oir.KCache) and c.fill
        }
        flushing_fields: Dict[str, str] = {
            c.name: filling_fields[c.name] if c.name in filling_fields else new_symbol_name(c.name)
            for c in node.caches
            if isinstance(c, oir.KCache) and c.flush
        }

        filling_or_flushing_fields = dict(
            set(filling_fields.items()) | set(flushing_fields.items())
        )

        if not filling_or_flushing_fields:
            return node

        # new temporaries used for caches, declarations are later added to stencil
        for field_name, tmp_name in filling_or_flushing_fields.items():
            new_tmps.append(
                oir.Temporary(
                    name=tmp_name, dtype=symtable[field_name].dtype, dimensions=(True, True, True)
                )
            )

        if filling_fields:
            # split sections where more than one fill operations are required at the entry level
            first_unfilled: Dict[str, int] = dict()
            split_sections: List[oir.VerticalLoopSection] = []
            for section in node.sections:
                split_section, previous_fills = self._split_section_with_multiple_fills(
                    node.loop_order, section, filling_fields, first_unfilled, new_symbol_name
                )
                split_sections += split_section
        else:
            split_sections = node.sections

        # generate cache fill and flush statements
        first_unfilled = dict()
        sections = []
        for section in split_sections:
            fills, first_unfilled = self._fill_stmts(
                node.loop_order, section, filling_fields, first_unfilled, symtable
            )
            flushes = self._flush_stmts(node.loop_order, section, flushing_fields, symtable)
            sections.append(
                self.visit(
                    section,
                    fills=fills,
                    flushes=flushes,
                    name_map=filling_or_flushing_fields,
                    symtable=symtable,
                    **kwargs,
                )
            )

        # replace cache declarations
        caches = [c for c in node.caches if c.name not in filling_or_flushing_fields] + [
            oir.KCache(name=f, fill=False, flush=False) for f in filling_or_flushing_fields.values()
        ]

        return oir.VerticalLoop(
            loop_order=node.loop_order,
            sections=sections,
            caches=caches,
            loc=node.loc,
        )

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        new_tmps: List[oir.Temporary] = []
        return oir.Stencil(
            name=node.name,
            params=node.params,
            vertical_loops=self.visit(
                node.vertical_loops,
                new_tmps=new_tmps,
                new_symbol_name=symbol_name_creator(set(kwargs["symtable"])),
                **kwargs,
            ),
            declarations=node.declarations + new_tmps,
            loc=node.loc,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, List, Optional, Set, Union

import dace

from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc import daceir as dcir
from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.definitions import Extent


if TYPE_CHECKING:
    from .nodes import StencilComputation

_EXPANSION_VALIDITY_CHECKS: List[Callable] = []


def _register_validity_check(x):
    _EXPANSION_VALIDITY_CHECKS.append(x)
    return x


@dataclass
class ExpansionItem:
    pass


@dataclass
class Iteration:
    axis: dcir.Axis
    kind: str  # tiling, contiguous
    # if stride is not specified, it is chosen based on backend (tiling) and loop order (K)
    stride: Optional[int] = None

    @property
    def iterations(self) -> List["Iteration"]:
        return [self]


@dataclass
class Map(ExpansionItem):
    iterations: List[Iteration]
    schedule: Optional[dace.ScheduleType] = None


@dataclass
class Loop(Iteration, ExpansionItem):
    kind: str = "contiguous"
    storage: dace.StorageType = None

    @property
    def iterations(self) -> List[Iteration]:
        return [self]


@dataclass
class Stages(ExpansionItem):
    pass


@dataclass
class Sections(ExpansionItem):
    pass


def _get_axis_from_pattern(item, fmt):
    for axis in dcir.Axis.dims_3d():
        if fmt.format(axis=axis) == item:
            return axis
    return ""


def _is_domain_loop(item):
    return _get_axis_from_pattern(item, fmt="{axis}Loop")


def _is_domain_map(item):
    return _get_axis_from_pattern(item, fmt="{axis}Map")


def _is_tiling(item):
    return _get_axis_from_pattern(item, fmt="Tile{axis}")


def get_expansion_order_axis(item):
    if axis := (
        _is_domain_map(item)
        or _is_domain_loop(item)
        or _is_tiling(item)
        or _get_axis_from_pattern(item, fmt="{axis}")
    ):
        return dcir.Axis(axis)
    raise ValueError(f"Can't get axis for item '{item}'.")


def get_expansion_order_index(expansion_order, axis):
    for idx, item in enumerate(expansion_order):
        if isinstance(item, Iteration) and item.axis == axis:
            return idx
        elif isinstance(item, Map):
            for it in item.iterations:
                if it.kind == "contiguous" and it.axis == axis:
                    return idx


def _is_expansion_order_implemented(expansion_specification):

    for item in expansion_specification:
        if isinstance(item, Sections):
            break
        if isinstance(item, Iteration) and item.axis == dcir.Axis.K:
            return False
        if isinstance(item, Map) and any(it.axis == dcir.Axis.K for it in item.iterations):
            return False

    return True


def _choose_loop_or_map(node, eo):
    if any(eo == axis for axis in dcir.Axis.dims_horizontal()):
        return f"{eo}Map"
    if eo == dcir.Axis.K:
        if node.oir_node.loop_order == common.LoopOrder.PARALLEL:
            return f"{eo}Map"
        else:
            return f"{eo}Loop"
    return eo


def _order_as_spec(computation_node, expansion_order):
    expansion_order = list(_choose_loop_or_map(computation_node, eo) for eo in expansion_order)
    expansion_specification = []
    for item in expansion_order:
        if isinstance(item, ExpansionItem):
            expansion_specification.append(item)
        elif axis := _is_tiling(item):
            expansion_specification.append(
                Map(
                    iterations=[
                        Iteration(
                            axis=axis,
                            kind="tiling",
                            stride=None,
                        )
                    ]
                )
            )
        elif axis := _is_domain_map(item):
            expansion_specification.append(
                Map(
                    iterations=[
                        Iteration(
                            axis=axis,
                            kind="contiguous",
                            stride=1,
                        )
                    ]
                )
            )
        elif axis := _is_domain_loop(item):
            expansion_specification.append(
                Loop(
                    axis=axis,
                    stride=-1
                    if computation_node.oir_node.loop_order == common.LoopOrder.BACKWARD
                    else 1,
                )
            )
        elif item == "Sections":
            expansion_specification.append(Sections())
        else:
            assert item == "Stages", item
            expansion_specification.append(Stages())

    return expansion_specification


def _populate_strides(node, expansion_specification):
    """Fill in `stride` attribute of `Iteration` and `Loop` dataclasses.

    For loops, stride is set to either -1 or 1, based on iteration order.
    For tiling maps, the stride is chosen such that the resulting tile size
    is that of the tile_size attribute.
    Other maps get stride 1.
    """
    assert all(isinstance(es, ExpansionItem) for es in expansion_specification)

    iterations = [it for item in expansion_specification for it in getattr(item, "iterations", [])]

    for it in iterations:
        if isinstance(it, Loop):
            if it.stride is None:
                if node.oir_node.loop_order == common.LoopOrder.BACKWARD:
                    it.stride = -1
                else:
                    it.stride = 1
        else:
            if it.stride is None:
                if it.kind == "tiling":
                    if node.extents is not None and it.axis.to_idx() < 2:
                        extent = Extent.zeros(2)
                        for he_extent in node.extents.values():
                            extent = extent.union(he_extent)
                        extent = extent[it.axis.to_idx()]
                    else:
                        extent = (0, 0)
                    it.stride = node.tile_strides.get(it.axis, 8)
                else:
                    it.stride = 1


def _populate_storages(self, expansion_specification):
    assert all(isinstance(es, ExpansionItem) for es in expansion_specification)
    innermost_axes = set(dcir.Axis.dims_3d())
    tiled_axes = set()
    for item in expansion_specification:
        if isinstance(item, Map):
            for it in item.iterations:
                if it.kind == "tiling":
                    tiled_axes.add(it.axis)
    for es in reversed(expansion_specification):
        if isinstance(es, Map):
            for it in es.iterations:
                if it.axis in innermost_axes:
                    innermost_axes.remove(it.axis)
                if it.kind == "tiling":
                    tiled_axes.remove(it.axis)


def _populate_cpu_schedules(self, expansion_specification):
    is_outermost = True
    for es in expansion_specification:
        if isinstance(es, Map):
            if es.schedule is None:
                if is_outermost:
                    es.schedule = dace.ScheduleType.CPU_Multicore
                    is_outermost = False
                else:
                    es.schedule = dace.ScheduleType.Default


def _populate_gpu_schedules(self, expansion_specification):
    # On GPU if any dimension is tiled and has a contiguous map in the same axis further in
    # pick those two maps as Device/ThreadBlock maps. If not, Make just device map with
    # default blocksizes
    is_outermost = True
    tiled = False
    for i, item in enumerate(expansion_specification):
        if isinstance(item, Map):
            for it in item.iterations:
                if not tiled and it.kind == "tiling":
                    for inner_item in expansion_specification[i + 1 :]:
                        if isinstance(inner_item, Map) and any(
                            inner_it.kind == "contiguous" and inner_it.axis == it.axis
                            for inner_it in inner_item.iterations
                        ):
                            item.schedule = dace.ScheduleType.GPU_Device
                            inner_item.schedule = dace.ScheduleType.GPU_ThreadBlock
                            tiled = True
                            break
    if not tiled:
        assert any(
            isinstance(item, Map) for item in expansion_specification
        ), "needs at least one map to avoid dereferencing on CPU"
        for es in expansion_specification:
            if isinstance(es, Map):
                if es.schedule is None:
                    if is_outermost:
                        es.schedule = dace.ScheduleType.GPU_Device
                        is_outermost = False
                    else:
                        es.schedule = dace.ScheduleType.Default


def _populate_schedules(self, expansion_specification):
    assert all(isinstance(es, ExpansionItem) for es in expansion_specification)
    assert hasattr(self, "_device")
    if self.device == dace.DeviceType.GPU:
        _populate_gpu_schedules(self, expansion_specification)
    else:
        _populate_cpu_schedules(self, expansion_specification)


def _collapse_maps_gpu(self, expansion_specification):
    def _union_map_items(last_item, next_item):
        if last_item.schedule == next_item.schedule:
            return (
                Map(
                    iterations=last_item.iterations + next_item.iterations,
                    schedule=last_item.schedule,
                ),
            )

        if next_item.schedule is None or next_item.schedule == dace.ScheduleType.Default:
            specified_item = last_item
        else:
            specified_item = next_item

        if specified_item.schedule is not None and not specified_item == dace.ScheduleType.Default:
            return (
                Map(
                    iterations=last_item.iterations + next_item.iterations,
                    schedule=specified_item.schedule,
                ),
            )

        # one is default and the other None
        return (
            Map(
                iterations=last_item.iterations + next_item.iterations,
                schedule=dace.ScheduleType.Default,
            ),
        )

    res_items = []
    for item in expansion_specification:
        if isinstance(item, Map):
            if not res_items or not isinstance(res_items[-1], Map):
                res_items.append(item)
            else:
                res_items[-1:] = _union_map_items(last_item=res_items[-1], next_item=item)
        else:
            res_items.append(item)
    for item in res_items:
        if isinstance(item, Map) and (
            item.schedule is None or item.schedule == dace.ScheduleType.Default
        ):
            item.schedule = dace.ScheduleType.Sequential
    return res_items


def _collapse_maps_cpu(self, expansion_specification):
    res_items = []
    for item in expansion_specification:
        if isinstance(item, Map):
            if (
                not res_items
                or not isinstance(res_items[-1], Map)
                or any(
                    it.axis in set(outer_it.axis for outer_it in res_items[-1].iterations)
                    for it in item.iterations
                )
            ):
                res_items.append(item)
            elif item.schedule == res_items[-1].schedule:
                res_items[-1].iterations.extend(item.iterations)
            elif item.schedule is None or item.schedule == dace.ScheduleType.Default:
                if res_items[-1].schedule == dace.ScheduleType.CPU_Multicore:
                    res_items[-1].iterations.extend(item.iterations)
                else:
                    res_items.append(item)
            elif (
                res_items[-1].schedule is None
                or res_items[-1].schedule == dace.ScheduleType.Default
            ):
                if item.schedule == dace.ScheduleType.CPU_Multicore:
                    res_items[-1].iterations.extend(item.iterations)
                    res_items[-1].schedule = dace.ScheduleType.CPU_Multicore
                else:
                    res_items.append(item)
            else:
                res_items.append(item)
        else:
            res_items.append(item)
    return res_items


def _collapse_maps(self, expansion_specification):
    assert hasattr(self, "_device")
    if self.device == dace.DeviceType.GPU:
        res_items = _collapse_maps_gpu(self, expansion_specification)
    else:
        res_items = _collapse_maps_cpu(self, expansion_specification)
    expansion_specification.clear()
    expansion_specification.extend(res_items)


def make_expansion_order(
    node: "StencilComputation", expansion_order: Union[List[str], List[ExpansionItem]]
) -> List[ExpansionItem]:
    if expansion_order is None:
        return None
    expansion_order = copy.deepcopy(expansion_order)
    expansion_specification = _order_as_spec(node, expansion_order)

    if not _is_expansion_order_implemented(expansion_specification):
        raise ValueError("Provided StencilComputation.expansion_order is not supported.")
    if node.oir_node is not None:
        if not is_expansion_order_valid(node, expansion_specification):
            raise ValueError("Provided StencilComputation.expansion_order is invalid.")

    _populate_strides(node, expansion_specification)
    _populate_schedules(node, expansion_specification)
    _collapse_maps(node, expansion_specification)
    _populate_storages(node, expansion_specification)
    return expansion_specification


def _k_inside_dims(node: "StencilComputation"):
    # Putting K inside of i or j is valid if
    # * K parallel or
    # * All reads with k-offset to values modified in same HorizontalExecution are not
    #   to fields that are also accessed horizontally (in I or J, respectively)
    #   (else, race condition in other column)

    if node.oir_node.loop_order == common.LoopOrder.PARALLEL:
        return {dcir.Axis.I, dcir.Axis.J}

    res = {dcir.Axis.I, dcir.Axis.J}
    for section in node.oir_node.sections:
        for he in section.horizontal_executions:
            i_offset_fields = set(
                (
                    acc.name
                    for acc in he.walk_values().if_isinstance(oir.FieldAccess)
                    if acc.offset.to_dict()["i"] != 0
                )
            )
            j_offset_fields = set(
                (
                    acc.name
                    for acc in he.walk_values().if_isinstance(oir.FieldAccess)
                    if acc.offset.to_dict()["j"] != 0
                )
            )
            k_offset_fields = set(
                (
                    acc.name
                    for acc in he.walk_values().if_isinstance(oir.FieldAccess)
                    if isinstance(acc.offset, oir.VariableKOffset) or acc.offset.to_dict()["k"] != 0
                )
            )
            modified_fields: Set[str] = (
                he.walk_values()
                .if_isinstance(oir.AssignStmt)
                .getattr("left")
                .if_isinstance(oir.FieldAccess)
                .getattr("name")
                .to_set()
            )
            for name in modified_fields:
                if name in k_offset_fields and name in i_offset_fields:
                    res.remove(dcir.Axis.I)
                if name in k_offset_fields and name in j_offset_fields:
                    res.remove(dcir.Axis.J)
    return res


def _k_inside_stages(node: "StencilComputation"):
    # Putting K inside of stages is valid if
    # * K parallel
    # * not "ahead" in order of iteration to fields that are modified in previous
    #   HorizontalExecutions (else, reading updated values that should be old)

    if node.oir_node.loop_order == common.LoopOrder.PARALLEL:
        return True

    for section in node.oir_node.sections:
        modified_fields: Set[str] = set()
        for he in section.horizontal_executions:
            if modified_fields:
                ahead_acc = list()
                for acc in he.walk_values().if_isinstance(oir.FieldAccess):
                    if (
                        isinstance(acc.offset, oir.VariableKOffset)
                        or (
                            node.oir_node.loop_order == common.LoopOrder.FORWARD
                            and acc.offset.k > 0
                        )
                        or (
                            node.oir_node.loop_order == common.LoopOrder.BACKWARD
                            and acc.offset.k < 0
                        )
                    ):
                        ahead_acc.append(acc)
                if any(acc.name in modified_fields for acc in ahead_acc):
                    return False

            modified_fields.update(
                he.walk_values()
                .if_isinstance(oir.AssignStmt)
                .getattr("left")
                .if_isinstance(oir.FieldAccess)
                .getattr("name")
                .to_set()
            )

    return True


@_register_validity_check
def _sequential_as_loops(
    node: "StencilComputation", expansion_specification: List[ExpansionItem]
) -> bool:
    # K can't be Map if not parallel
    if node.oir_node.loop_order != common.LoopOrder.PARALLEL and any(
        (isinstance(item, Map) and any(it.axis == dcir.Axis.K for it in item.iterations))
        for item in expansion_specification
    ):
        return False
    return True


@_register_validity_check
def _stages_inside_sections(expansion_specification: List[ExpansionItem], **kwargs) -> bool:
    # Oir defines that HorizontalExecutions have to be applied per VerticalLoopSection. A meaningful inversion of this
    # is not possible.
    sections_idx = next(
        idx for idx, item in enumerate(expansion_specification) if isinstance(item, Sections)
    )
    stages_idx = next(
        idx for idx, item in enumerate(expansion_specification) if isinstance(item, Stages)
    )
    if stages_idx < sections_idx:
        return False
    return True


@_register_validity_check
def _k_inside_ij_valid(
    node: "StencilComputation", expansion_specification: List[ExpansionItem]
) -> bool:
    # OIR defines that horizontal maps go inside vertical K loop (i.e. all grid points are updated in a
    # HorizontalExecution before the computation of the next one is executed.).  Under certain conditions the semantics
    # remain unchanged even if a single horizontal map is executing all contained HorizontalExecution nodes.
    # Note: Opposed to e.g. Fusions in OIR, this can here be done on a per-dimension basis. See `_k_inside_dims` for
    # details.
    for axis in dcir.Axis.dims_horizontal():
        if get_expansion_order_index(expansion_specification, axis) < get_expansion_order_index(
            expansion_specification, dcir.Axis.K
        ) and axis not in _k_inside_dims(node):
            return False
    return True


@_register_validity_check
def _k_inside_stages_valid(
    node: "StencilComputation", expansion_specification: List[ExpansionItem]
) -> bool:
    # OIR defines that all horizontal executions of a VerticalLoopSection are run per level. Under certain conditions
    # the semantics remain unchanged even if the k loop is run per horizontal execution. See `_k_inside_stages` for
    # details
    stages_idx = next(
        idx for idx, item in enumerate(expansion_specification) if isinstance(item, Stages)
    )
    if stages_idx < get_expansion_order_index(
        expansion_specification, dcir.Axis.K
    ) and not _k_inside_stages(node):
        return False
    return True


@_register_validity_check
def _ij_outside_sections_valid(
    node: "StencilComputation", expansion_specification: List[ExpansionItem]
) -> bool:
    # If there are multiple horizontal executions in any section, IJ iteration must go inside sections.
    # TODO: do mergeability checks on a per-axis basis.
    for item in expansion_specification:
        if isinstance(item, Sections):
            break
        if isinstance(item, (Map, Loop, Iteration)):
            for it in item.iterations:
                if it.axis in dcir.Axis.dims_horizontal() and it.kind == "contiguous":
                    if any(
                        len(section.horizontal_executions) > 1 for section in node.oir_node.sections
                    ):
                        return False

    # if there are horizontal executions with different iteration ranges in an axis across sections,
    # that iteration must be per section
    # TODO less conservative: allow different domains if all outputs smaller than bounding box are temporaries
    # TODO implement/allow this with predicates implicit regions / predicates
    for item in expansion_specification:
        if isinstance(item, Sections):
            break
        for it in getattr(item, "iterations", []):
            if it.axis in dcir.Axis.dims_horizontal() and it.kind == "contiguous":
                xiter = iter(node.oir_node.walk_values().if_isinstance(oir.HorizontalExecution))
                extent = node.get_extents(next(xiter))
                for he in xiter:
                    if node.get_extents(he)[it.axis.to_idx()] != extent[it.axis.to_idx()]:
                        return False
    return True


@_register_validity_check
def _iterates_domain(expansion_specification: List[ExpansionItem], **kwargs) -> bool:
    # There must be exactly one iteration per dimension, except for tiled dimensions, where a Tiling has to go outside
    # and the corresponding contiguous iteration inside.
    tiled_axes = set()
    contiguous_axes = set()
    for item in expansion_specification:
        if isinstance(item, (Map, Loop, Iteration)):
            for it in item.iterations:
                if it.kind == "tiling":
                    if it.axis in tiled_axes or it.axis in contiguous_axes:
                        return False
                    tiled_axes.add(it.axis)
                else:
                    if it.axis in contiguous_axes:
                        return False
                    contiguous_axes.add(it.axis)
    if not all(axis in contiguous_axes for axis in dcir.Axis.dims_3d()):
        return False
    return True


def is_expansion_order_valid(node: "StencilComputation", expansion_order) -> bool:
    """Check if a given expansion specification valid.

    That is, it is semantically valid for the StencilComputation node that is to be configured and currently
    implemented.
    """
    expansion_specification = list(_choose_loop_or_map(node, eo) for eo in expansion_order)

    for check in _EXPANSION_VALIDITY_CHECKS:
        if not check(node=node, expansion_specification=expansion_specification):
            return False

    return _is_expansion_order_implemented(expansion_specification)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import collections
import inspect
import sys
import types
from itertools import count, product

import hypothesis as hyp
import hypothesis.strategies as hyp_st
import numpy as np
import pytest

import gt4py.cartesian.gtc.utils as gtc_utils
from gt4py import cartesian as gt
from gt4py import cartesian as gt4pyc
from gt4py import storage as gt_storage
from gt4py.cartesian import gtscript
from gt4py.cartesian import utils as gt_utils
from gt4py.cartesian.definitions import AccessKind, Boundary, CartesianSpace, FieldInfo
from gt4py.cartesian.frontend.nodes import Index
from gt4py.cartesian.gtc.definitions import Shape
from gt4py.cartesian.stencil_object import StencilObject
from gt4py.storage import utils as storage_utils

from .input_strategies import (
    SymbolKind,
    composite_implementation_strategy_factory,
    composite_strategy_factory,
    derived_shape_st,
    ndarray_shape_st,
    ndarray_st,
)
from .utils import annotate_function, standardize_dtype_dict


ParameterSet = type(pytest.param())

counter = count()
RTOL = 1e-05
ATOL = 1e-08
EQUAL_NAN = False

unique_str_ctr = count()


def unique_str():
    return str(next(unique_str_ctr))


class SuiteMeta(type):
    """Custom metaclass for all :class:`StencilTestSuite` classes.

    This metaclass generates a namespace-like class where all methods are static
    and adds new members with the required testing strategies and expected results
    to test a stencil definition and implementations using Hypothesis and pytest.
    """

    required_members = {"domain_range", "symbols", "definition", "validation", "backends", "dtypes"}

    def collect_symbols(cls_name, cls_dict):
        domain_range = cls_dict["domain_range"]
        domain_strategy = cls_dict["domain_strategy"] = hyp_st.shared(
            ndarray_shape_st(domain_range), key=cls_name
        )

        generation_strategy_factories = cls_dict["generation_strategy_factories"] = dict()
        implementation_strategy_factories = cls_dict["implementation_strategy_factories"] = dict()
        global_boundaries = cls_dict["global_boundaries"] = dict()
        constants = cls_dict["constants"] = dict()
        singletons = cls_dict["singletons"] = dict()
        cls_dict["field_params"] = field_params = {}
        max_boundary = ((0, 0), (0, 0), (0, 0))
        for symbol in cls_dict["symbols"].values():
            if symbol.kind == SymbolKind.FIELD:
                max_boundary = tuple(
                    (max(m[0], b[0]), max(m[1], b[1]))
                    for m, b in zip(max_boundary, symbol.boundary)
                )
        cls_dict["max_boundary"] = max_boundary

        for name, symbol in cls_dict["symbols"].items():
            if symbol.kind == SymbolKind.GLOBAL_STRATEGY:
                generation_strategy_factories[name] = symbol.value_st_factory
            elif symbol.kind == SymbolKind.GLOBAL_SET:
                constants[name] = symbol.values
            elif symbol.kind == SymbolKind.SINGLETON:
                singletons[name] = symbol.values[0]
            elif symbol.kind == SymbolKind.FIELD:
                if symbol.axes:
                    axes = symbol.axes
                    extra_shape = tuple(
                        b[0] + b[1] if ax in symbol.axes else None
                        for b, ax in zip(max_boundary, "IJK")
                    )
                else:
                    axes = "IJK"
                    extra_shape = tuple(b[0] + b[1] for b in max_boundary)

                if symbol.data_dims:
                    data_dims = symbol.data_dims
                    extra_shape = (*extra_shape, *symbol.data_dims)
                else:
                    data_dims = tuple()

                field_params[name] = (axes, data_dims)

                global_boundaries[name] = symbol.boundary
                shape_strategy = derived_shape_st(domain_strategy, extra_shape)

                # Use default arguments to pass values avoiding late binding problems
                def implementation_strategy_factory(
                    dt, shape=shape_strategy, value_st_factory=symbol.value_st_factory
                ):
                    return ndarray_st(dt, shape, value_st_factory)

                implementation_strategy_factories[name] = implementation_strategy_factory
            elif symbol.kind == SymbolKind.PARAMETER:
                implementation_strategy_factories[name] = symbol.value_st_factory
            elif symbol.kind == SymbolKind.NONE:
                implementation_strategy_factories[name] = symbol.value_st_factory

            else:
                raise AssertionError

        cls_dict["origin"] = tuple(o[0] for o in max_boundary)

    def parametrize_generation_tests(cls_name, cls_dict):

        backends = cls_dict["backends"]
        dtypes = cls_dict["dtypes"]
        field_params = cls_dict["field_params"]
        generation_strategy_factories = cls_dict["generation_strategy_factories"]

        def get_dtype_combinations(dtypes):
            grouped_combinations = [
                {k: v for k, v in zip(dtypes.keys(), p)} for p in product(*dtypes.values())
            ]
            ret = [
                {k: combination[ktuple] for ktuple in combination for k in ktuple}
                for combination in grouped_combinations
            ]
            return ret

        def get_globals_combinations(dtypes):
            combinations = [
                {k: dtypes[k].type(v) for k, v in zip(cls_dict["constants"].keys(), p)}
                for p in product(*cls_dict["constants"].values())
            ]
            if not combinations:
                return [{}]
            else:
                return combinations

        cls_dict["tests"] = []
        for b in backends:
            for d in get_dtype_combinations(dtypes):
                for g in get_globals_combinations(d):
                    cls_dict["tests"].append(
                        dict(
                            backend=b if isinstance(b, str) else b.values[0],
                            marks=[] if isinstance(b, str) else b.marks,
                            suite=cls_name,
                            constants=g,
                            dtypes=d,
                            generation_strategy=composite_strategy_factory(
                                d, generation_strategy_factories
                            ),
                            implementations=[],
                            test_id=len(cls_dict["tests"]),
                            definition=annotate_function(
                                function=cls_dict["definition"],
                                dtypes={
                                    name: (
                                        dtype.type
                                        if (name not in field_params)
                                        else gtscript.Field[
                                            getattr(gtscript, field_params[name][0]),
                                            (dtype.type, field_params[name][1]),
                                        ]
                                    )
                                    for name, dtype in d.items()
                                },
                            ),
                        )
                    )

        def generation_test_wrapper(self, test):
            @hyp.given(hypothesis_data=test["generation_strategy"]())
            def hyp_wrapper(test_hyp, hypothesis_data):
                self._test_generation(
                    test_hyp, {**test_hyp["constants"], **hypothesis_data, **cls_dict["singletons"]}
                )

            hyp_wrapper(test)

        pytest_params = []

        for test in cls_dict["tests"]:
            if test["suite"] == cls_name:
                marks = test["marks"]
                if gt4pyc.backend.from_name(test["backend"]).storage_info["device"] == "gpu":
                    marks.append(pytest.mark.requires_gpu)
                name = test["backend"]
                name += "".join(f"_{key}_{value}" for key, value in test["constants"].items())
                name += "".join(
                    "_{}_{}".format(key, value.name) for key, value in test["dtypes"].items()
                )
                param = pytest.param(test, marks=marks, id=name)
                pytest_params.append(param)

        cls_dict["test_generation"] = pytest.mark.parametrize("test", pytest_params)(
            generation_test_wrapper
        )

    def parametrize_implementation_tests(cls_name, cls_dict):
        implementation_strategy_factories = cls_dict["implementation_strategy_factories"]
        global_boundaries = cls_dict["global_boundaries"]

        def implementation_test_wrapper(self, test, implementation_strategy):
            @hyp.given(hypothesis_data=implementation_strategy())
            def hyp_wrapper(test_hyp, hypothesis_data):
                self._test_implementation(test_hyp, hypothesis_data)

            hyp_wrapper(test)

        runtime_pytest_params = []
        for test in cls_dict["tests"]:
            if test["suite"] == cls_name:
                marks = test["marks"]
                if gt4pyc.backend.from_name(test["backend"]).storage_info["device"] == "gpu":
                    marks.append(pytest.mark.requires_gpu)
                name = test["backend"]
                name += "".join(f"_{key}_{value}" for key, value in test["constants"].items())
                name += "".join(
                    "_{}_{}".format(key, value.name) for key, value in test["dtypes"].items()
                )
                runtime_pytest_params.append(
                    pytest.param(
                        test,
                        composite_implementation_strategy_factory(
                            test["dtypes"], implementation_strategy_factories, global_boundaries
                        ),
                        marks=marks,
                        id=name,
                    )
                )

        cls_dict["test_implementation"] = pytest.mark.parametrize(
            ("test", "implementation_strategy"), runtime_pytest_params
        )(implementation_test_wrapper)

    @classmethod
    def _validate_new_args(cls, cls_name, cls_dict):
        missing_members = cls.required_members - cls_dict.keys()
        if len(missing_members) > 0:
            raise TypeError(
                "Missing {missing} required members in '{name}' definition".format(
                    missing=missing_members, name=cls_name
                )
            )
        # Check class dict
        domain_range = cls_dict["domain_range"]
        backends = cls_dict["backends"]

        # Create testing strategies
        assert isinstance(cls_dict["symbols"], collections.abc.Mapping), "Invalid 'symbols' mapping"

        # Check domain and ndims
        assert 1 <= len(domain_range) <= 3 and all(
            len(d) == 2 for d in domain_range
        ), "Invalid 'domain_range' definition"

        if any(cls_name.endswith(suffix) for suffix in ("1D", "2D", "3D")):
            assert cls_dict["ndims"] == int(
                cls_name[-2:-1]
            ), "Suite name does not match the actual 'ndims'"

        # Check dtypes
        assert isinstance(
            cls_dict["dtypes"], (collections.abc.Sequence, collections.abc.Mapping)
        ), "'dtypes' must be a sequence or a mapping object"

        # Check backends
        if not all(
            isinstance(b, str)
            or (isinstance(b, ParameterSet) and len(b.values) == 1 and isinstance(b.values[0], str))
            for b in backends
        ):
            raise TypeError("'backends' must be a sequence of strings")
        backends = [pytest.param(b) if isinstance(b, str) else b for b in backends]
        for b in backends:
            if b.values[0] not in gt.backend.REGISTRY.names:
                raise ValueError("backend '{backend}' not supported".format(backend=b))

        # Check definition and validation functions
        if not isinstance(cls_dict["definition"], types.FunctionType):
            raise TypeError("The 'definition' attribute must be a stencil definition function")
        if not isinstance(cls_dict["validation"], types.FunctionType):
            raise TypeError("The 'validation' attribute must be a validation function")

    def __new__(cls, cls_name, bases, cls_dict):
        if cls_dict.get("_skip_", False):  # skip metaclass magic
            return super().__new__(cls, cls_name, bases, cls_dict)
        # Grab members inherited from base classes

        missing_members = cls.required_members - cls_dict.keys()

        for key in missing_members:
            for base in bases:
                if hasattr(base, key):
                    cls_dict[key] = getattr(base, key)
                    break

        dtypes = cls_dict["dtypes"]
        if isinstance(dtypes, collections.abc.Sequence):
            dtypes = {tuple(cls_dict["symbols"].keys()): dtypes}
        cls_dict["dtypes"] = standardize_dtype_dict(dtypes)
        cls_dict["ndims"] = len(cls_dict["domain_range"])

        # Filter out unsupported backends
        cls_dict["backends"] = [
            backend
            for backend in cls_dict["backends"]
            if gt4pyc.backend.from_name(backend if isinstance(backend, str) else backend.values[0])
            is not None
        ]

        cls._validate_new_args(cls_name, cls_dict)

        # Extract input and parameter names
        input_names = []
        parameter_names = []
        definition_signature = inspect.signature(cls_dict["definition"])
        validation_signature = inspect.signature(cls_dict["validation"])
        for (def_name, def_pobj), (val_name, val_pobj) in zip(
            definition_signature.parameters.items(), validation_signature.parameters.items()
        ):
            if def_name != val_name or def_pobj.kind != val_pobj.kind:
                raise ValueError(
                    "Incompatible signatures for 'definition' and 'validation' functions"
                )

            if def_pobj.kind == inspect.Parameter.KEYWORD_ONLY:
                parameter_names.append(def_name)
                if def_pobj.default != inspect.Parameter.empty:
                    assert def_pobj.default == val_pobj.default
            else:
                input_names.append(def_name)

        cls.collect_symbols(cls_name, cls_dict)

        assert set(input_names + parameter_names) == set(
            cls_dict["implementation_strategy_factories"].keys()
        ), "Missing or invalid keys in 'symbols' mapping (generated: {})".format(
            cls_dict["implementation_strategy_factories"].keys()
        )

        cls.parametrize_generation_tests(cls_name, cls_dict)
        cls.parametrize_implementation_tests(cls_name, cls_dict)

        return super().__new__(cls, cls_name, bases, cls_dict)


class StencilTestSuite(metaclass=SuiteMeta):
    """Base class for every *stencil test suite*.

    Every new test suite must inherit from this class and define proper
    attributes and methods to generate a valid set of testing strategies.
    For compatibility with pytest, suites must have names starting with 'Test'

    Supported and required class attributes are:

    Attributes
    ----------
    dtypes : `dict` or `list`
        Required class attribute.
        GlobalDecl suite dtypes dictionary.
        - ``label``: `list` of `dtype`.
        If this value is a `list`, it will be converted to a `dict` with the default
        `None` key assigned to its value. It is meant to be populated with labels representing
        groups of symbols that should have the same type.

        Example:

        .. code-block:: python

                    {
                        'float_symbols' : (np.float32, np.float64),
                        'int_symbols' : (int, np.int_, np.int64)
                    }

    domain_range : `Sequence` of pairs like `((int, int), (int, int) ... )`
        Required class attribute.
        CartesianSpace sizes for testing. Each item encodes the (min, max) range of sizes for every axis.
    symbols : `dict`
        Required class attribute.
        Definition of symbols (globals, parameters and inputs) used in this stencil.
        - ``name``: `utils._SymbolDescriptor`. It is recommended to use the convenience
        functions `global_name()`, `parameter()` and `field()`. These functions have similar
        and self-explanatory arguments. Note that `dtypes` could be a list of actual dtypes
        but it is usually assumed to be a label from the global suite dtypes dictionary.
    definition : `function`
        Required class attribute.
        Stencil definition function.
    validation : `function`
        Required class attribute.
        Stencil validation function. It should have exactly the same signature than
        arguments to access the actual values used in the current testing invocation.
        the ``definition`` function plus the extra ``_globals_``, ``_domain_``, and ``_origin_``
        It should always return a `list` of `numpy.ndarray` s, one per output, even if
        the function only defines one output value.
    definition_strategies : `dict`
        Automatically generated.
        Hypothesis strategies for the stencil parameters used at definition (externals)
        - ``constant_name``: Hypothesis strategy (`strategy`).
    validation_strategies : `dict`
        Automatically generated.
        Hypothesis strategies for the stencil parameters used at run-time (fields and parameters)
        - ``field_name``: Hypothesis strategy (`strategy`).
        - ``parameter_name``: Hypothesis strategy (`strategy`).
    ndims : `int`
        Automatically generated.
        Constant of dimensions (1-3). If the name of class ends in ["1D", "2D", "3D"],
        this attribute needs to match the name or an assertion error will be raised.
    global_boundaries : `dict`
        Automatically generated.
        Expected global boundaries for the input fields.
        - ``field_name``: 'list' of ``ndim`` 'tuple`s  (``(lower_boundary, upper_boundary)``).
        Example (3D): `[(1, 3), (2, 2), (0, 0)]`
    """

    _skip_ = True  # Avoid processing of this empty test suite

    @classmethod
    def _test_generation(cls, test, externals_dict):
        """Test source code generation for all *backends* and *stencil suites*.

        The generated implementations are cached in a :class:`utils.ImplementationsDB`
        instance, to avoid duplication of (potentially expensive) compilations.
        """
        backend_slug = gt_utils.slugify(test["backend"], valid_symbols="")
        implementation = gtscript.stencil(
            backend=test["backend"],
            definition=test["definition"],
            name=cls.__module__ + f".{test['suite']}_{backend_slug}_{test['test_id']}",
            rebuild=True,
            externals=externals_dict,
        )

        for k, v in externals_dict.items():
            implementation.constants[k] = v

        assert isinstance(implementation, StencilObject)
        assert implementation.backend == test["backend"]

        for name, field_info in implementation.field_info.items():
            if field_info.access == AccessKind.NONE:
                continue
            for i, ax in enumerate("IJK"):
                assert (
                    ax not in field_info.axes
                    or ax == "K"
                    or field_info.boundary[i] >= cls.global_boundaries[name][i]
                )
        test["implementations"].append(implementation)

    @classmethod
    def _run_test_implementation(cls, parameters_dict, implementation):  # noqa: C901  # too complex
        input_data, exec_info = parameters_dict

        origin = cls.origin
        max_boundary = Boundary(cls.max_boundary)
        field_params = cls.field_params
        field_dimensions = {}
        field_masks = {}
        for name, value in input_data.items():
            if isinstance(value, np.ndarray):
                field_masks[name] = tuple(
                    ax in field_params[name][0] for ax in CartesianSpace.names
                )
                field_dimensions[name] = tuple(
                    ax for ax in CartesianSpace.names if ax in field_params[name][0]
                )

        data_shape = Shape((sys.maxsize,) * 3)
        for name, data in input_data.items():
            if isinstance(data, np.ndarray):
                data_shape &= Shape(
                    gtc_utils.interpolate_mask(data.shape, field_masks[name], default=sys.maxsize)
                )

        domain = data_shape - (
            Index(max_boundary.lower_indices) + Index(max_boundary.upper_indices)
        )

        referenced_inputs = {
            name: info for name, info in implementation.field_info.items() if info is not None
        }
        referenced_inputs.update(
            {name: info for name, info in implementation.parameter_info.items() if info is not None}
        )

        # set externals for validation method
        for k, v in implementation.constants.items():
            sys.modules[cls.__module__].__dict__[k] = v

        # copy input data
        test_values = {}
        validation_values = {}
        for name, data in input_data.items():
            data = input_data[name]
            if name in referenced_inputs:
                info = referenced_inputs[name]
                if isinstance(info, FieldInfo):
                    data_dims = field_params[name][1]
                    if data_dims:
                        dtype = (data.dtype, data_dims)
                    else:
                        dtype = data.dtype
                    test_values[name] = gt_storage.from_array(
                        data=data,
                        dtype=dtype,
                        dimensions=field_dimensions[name],
                        aligned_index=gtc_utils.filter_mask(origin, field_masks[name]),
                        backend=implementation.backend,
                    )
                    validation_values[name] = np.array(data)
                else:
                    test_values[name] = data
                    validation_values[name] = data
            else:
                test_values[name] = None
                validation_values[name] = None

        # call implementation
        implementation(**test_values, origin=origin, domain=domain, exec_info=exec_info)

        # for validation data, data is cropped to actually touched domain, so that origin offseting
        # does not have to be implemented for every test suite. This is done based on info
        # specified in test suite
        cropped_validation_values = {}
        for name, data in validation_values.items():
            sym = cls.symbols[name]
            if data is not None and sym.kind == SymbolKind.FIELD:
                field_extent_low = tuple(b[0] for b in sym.boundary)
                offset_low = tuple(b[0] - e for b, e in zip(max_boundary, field_extent_low))
                field_extent_high = tuple(b[1] for b in sym.boundary)
                offset_high = tuple(b[1] - e for b, e in zip(max_boundary, field_extent_high))
                validation_slice = gtc_utils.filter_mask(
                    tuple(slice(o, s - h) for o, s, h in zip(offset_low, data_shape, offset_high)),
                    field_masks[name],
                )
                data_dims = field_params[name][1]
                if data_dims:
                    validation_slice = tuple([*validation_slice] + [slice(None)] * len(data_dims))
                cropped_validation_values[name] = data[validation_slice]
            else:
                cropped_validation_values[name] = data

        cls.validation(
            **cropped_validation_values,
            domain=domain,
            origin={
                name: info.boundary.lower_indices
                for name, info in implementation.field_info.items()
                if info is not None
            },
        )

        # Test values
        for name, value in test_values.items():
            if isinstance(value, np.ndarray):
                expected_value = validation_values[name]
                value = storage_utils.cpu_copy(value)

                np.testing.assert_allclose(
                    value,
                    expected_value,
                    rtol=RTOL,
                    atol=ATOL,
                    equal_nan=EQUAL_NAN,
                    err_msg="Wrong data in output field '{name}'".format(name=name),
                )

    @classmethod
    def _test_implementation(cls, test, parameters_dict):
        """Test computed values for implementations generated for all *backends* and *stencil suites*.

        The generated implementations are reused from previous tests by means of a
        :class:`utils.ImplementationsDB` instance shared at module scope.
        """
        implementation_list = test["implementations"]
        if not implementation_list:
            pytest.skip(
                "Cannot perform validation tests, since there are no valid implementations."
            )
        for implementation in implementation_list:
            if not isinstance(implementation, StencilObject):
                raise RuntimeError("Wrong function got from implementations_db cache!")

            cls._run_test_implementation(parameters_dict, implementation)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import pytest

from gt4py.eve import exceptions


def test_exception_types():
    with pytest.raises(TypeError):
        raise exceptions.EveTypeError()

    with pytest.raises(ValueError):
        raise exceptions.EveValueError()

    with pytest.raises(RuntimeError):
        raise exceptions.EveRuntimeError()


def test_default_exception_message():
    with pytest.raises(TypeError, match="my_data=23, your_data=42"):
        raise exceptions.EveTypeError(my_data=23, your_data=42)

    with pytest.raises(ValueError, match="my_data=23, your_data=42"):
        raise exceptions.EveValueError(my_data=23, your_data=42)

    with pytest.raises(RuntimeError, match="my_data=23, your_data=42"):
        raise exceptions.EveRuntimeError(my_data=23, your_data=42)


def test_custom_exception_message():
    with pytest.raises(TypeError, match="custom message"):
        raise exceptions.EveTypeError("This is a custom message", my_data=23, your_data=42)

    with pytest.raises(ValueError, match="custom message"):
        raise exceptions.EveValueError("This is a custom message", my_data=23, your_data=42)

    with pytest.raises(RuntimeError, match="custom message"):
        raise exceptions.EveRuntimeError("This is a custom message", my_data=23, your_data=42)
